import pytest
from flask import json

from authlib.oauth2.rfc7592 import (
    ClientConfigurationEndpoint as _ClientConfigurationEndpoint,
)

from .models import Client
from .models import Token
from .models import db


class ClientConfigurationEndpoint(_ClientConfigurationEndpoint):
    software_statement_alg_values_supported = ["RS256"]

    def authenticate_token(self, request):
        auth_header = request.headers.get("Authorization")
        if auth_header:
            access_token = auth_header.split()[1]
            return Token.query.filter_by(access_token=access_token).first()

    def update_client(self, client, client_metadata, request):
        client.set_client_metadata(client_metadata)
        db.session.add(client)
        db.session.commit()
        return client

    def authenticate_client(self, request):
        client_id = request.uri.split("/")[-1]
        return Client.query.filter_by(client_id=client_id).first()

    def revoke_access_token(self, request, token):
        token.revoked = True
        db.session.add(token)
        db.session.commit()

    def check_permission(self, client, request):
        client_id = request.uri.split("/")[-1]
        return client_id != "unauthorized_client_id"

    def delete_client(self, client, request):
        db.session.delete(client)
        db.session.commit()

    def generate_client_registration_info(self, client, request):
        return {
            "registration_client_uri": request.uri,
            "registration_access_token": request.headers["Authorization"].split(" ")[1],
        }


@pytest.fixture
def metadata():
    return {}


@pytest.fixture(autouse=True)
def server(server, app, metadata):
    @app.route("/configure_client/<client_id>", methods=["PUT", "GET", "DELETE"])
    def configure_client(client_id):
        return server.create_endpoint_response(
            ClientConfigurationEndpoint.ENDPOINT_NAME
        )

    class MyClientConfiguration(ClientConfigurationEndpoint):
        def get_server_metadata(test_client):
            return metadata

    server.register_endpoint(MyClientConfiguration)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "client_name": "Authlib",
            "scope": "openid profile",
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_read_client(test_client, client, token):
    assert client.client_name == "Authlib"
    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.get("/configure_client/client-id", headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 200
    assert resp["client_id"] == client.client_id
    assert resp["client_name"] == "Authlib"
    assert (
        resp["registration_client_uri"] == "http://localhost/configure_client/client-id"
    )
    assert resp["registration_access_token"] == token.access_token


def test_read_access_denied(test_client):
    rv = test_client.get("/configure_client/client-id")
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"

    headers = {"Authorization": "bearer invalid_token"}
    rv = test_client.get("/configure_client/client-id", headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"

    headers = {"Authorization": "bearer unauthorized_token"}
    rv = test_client.get(
        "/configure_client/client-id",
        json={"client_id": "client-id", "client_name": "new client_name"},
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"


def test_read_invalid_client(test_client, token):
    # If the client does not exist on this server, the server MUST respond
    # with HTTP 401 Unauthorized, and the registration access token used to
    # make this request SHOULD be immediately revoked.

    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.get("/configure_client/invalid_client_id", headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 401
    assert resp["error"] == "invalid_client"


def test_read_unauthorized_client(test_client, token):
    # If the client does not have permission to read its record, the server
    # MUST return an HTTP 403 Forbidden.

    client = Client(
        client_id="unauthorized_client_id",
        client_secret="unauthorized_client_secret",
    )
    db.session.add(client)

    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.get("/configure_client/unauthorized_client_id", headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 403
    assert resp["error"] == "unauthorized_client"


def test_update_client(test_client, client, token):
    # Valid values of client metadata fields in this request MUST replace,
    # not augment, the values previously associated with this client.
    # Omitted fields MUST be treated as null or empty values by the server,
    # indicating the client's request to delete them from the client's
    # registration.  The authorization server MAY ignore any null or empty
    # value in the request just as any other value.

    assert client.client_name == "Authlib"
    headers = {"Authorization": f"bearer {token.access_token}"}
    body = {
        "client_id": client.client_id,
        "client_name": "NewAuthlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 200
    assert resp["client_id"] == client.client_id
    assert resp["client_name"] == "NewAuthlib"
    assert client.client_name == "NewAuthlib"
    assert client.scope == ""


def test_update_access_denied(test_client):
    rv = test_client.put("/configure_client/client-id", json={})
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"

    headers = {"Authorization": "bearer invalid_token"}
    rv = test_client.put("/configure_client/client-id", json={}, headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"

    headers = {"Authorization": "bearer unauthorized_token"}
    rv = test_client.put(
        "/configure_client/client-id",
        json={"client_id": "client-id", "client_name": "new client_name"},
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"


def test_update_invalid_request(test_client, token):
    headers = {"Authorization": f"bearer {token.access_token}"}

    # The client MUST include its 'client_id' field in the request...
    rv = test_client.put("/configure_client/client-id", json={}, headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "invalid_request"

    # ... and it MUST be the same as its currently issued client identifier.
    rv = test_client.put(
        "/configure_client/client-id",
        json={"client_id": "invalid_client_id"},
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "invalid_request"

    # The updated client metadata fields request MUST NOT include the
    # 'registration_access_token', 'registration_client_uri',
    # 'client_secret_expires_at', or 'client_id_issued_at' fields
    rv = test_client.put(
        "/configure_client/client-id",
        json={
            "client_id": "client-id",
            "registration_client_uri": "https://client.test",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "invalid_request"

    # If the client includes the 'client_secret' field in the request,
    # the value of this field MUST match the currently issued client
    # secret for that client.
    rv = test_client.put(
        "/configure_client/client-id",
        json={"client_id": "client-id", "client_secret": "invalid_secret"},
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "invalid_request"


def test_update_invalid_client(test_client, token):
    # If the client does not exist on this server, the server MUST respond
    # with HTTP 401 Unauthorized, and the registration access token used to
    # make this request SHOULD be immediately revoked.

    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.put(
        "/configure_client/invalid_client_id",
        json={"client_id": "invalid_client_id", "client_name": "new client_name"},
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 401
    assert resp["error"] == "invalid_client"


def test_update_unauthorized_client(test_client, token):
    # If the client does not have permission to read its record, the server
    # MUST return an HTTP 403 Forbidden.

    client = Client(
        client_id="unauthorized_client_id",
        client_secret="unauthorized_client_secret",
    )
    db.session.add(client)

    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.put(
        "/configure_client/unauthorized_client_id",
        json={
            "client_id": "unauthorized_client_id",
            "client_name": "new client_name",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 403
    assert resp["error"] == "unauthorized_client"


def test_update_invalid_metadata(test_client, metadata, client, token):
    metadata["token_endpoint_auth_methods_supported"] = ["client_secret_basic"]
    headers = {"Authorization": f"bearer {token.access_token}"}

    # For all metadata fields, the authorization server MAY replace any
    # invalid values with suitable default values, and it MUST return any
    # such fields to the client in the response.
    # If the client attempts to set an invalid metadata field and the
    # authorization server does not set a default value, the authorization
    # server responds with an error as described in [RFC7591].

    body = {
        "client_id": client.client_id,
        "client_name": "NewAuthlib",
        "token_endpoint_auth_method": "invalid_auth_method",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "invalid_client_metadata"


def test_update_scopes_supported(test_client, metadata, token):
    metadata["scopes_supported"] = ["profile", "email"]

    headers = {"Authorization": f"bearer {token.access_token}"}
    body = {
        "client_id": "client-id",
        "scope": "profile email",
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["client_id"] == "client-id"
    assert resp["client_name"] == "Authlib"
    assert resp["scope"] == "profile email"

    headers = {"Authorization": f"bearer {token.access_token}"}
    body = {
        "client_id": "client-id",
        "scope": "",
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["client_id"] == "client-id"
    assert resp["client_name"] == "Authlib"

    body = {
        "client_id": "client-id",
        "scope": "profile email address",
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_update_response_types_supported(test_client, metadata, token):
    metadata["response_types_supported"] = ["code"]

    headers = {"Authorization": f"bearer {token.access_token}"}
    body = {
        "client_id": "client-id",
        "response_types": ["code"],
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["client_id"] == "client-id"
    assert resp["client_name"] == "Authlib"
    assert resp["response_types"] == ["code"]

    # https://datatracker.ietf.org/doc/html/rfc7592#section-2.2
    # If omitted, the default is that the client will use only the "code"
    # response type.
    body = {"client_id": "client-id", "client_name": "Authlib"}
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert "response_types" not in resp

    body = {
        "client_id": "client-id",
        "response_types": ["code", "token"],
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_update_grant_types_supported(test_client, metadata, token):
    metadata["grant_types_supported"] = ["authorization_code", "password"]

    headers = {"Authorization": f"bearer {token.access_token}"}
    body = {
        "client_id": "client-id",
        "grant_types": ["password"],
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["client_id"] == "client-id"
    assert resp["client_name"] == "Authlib"
    assert resp["grant_types"] == ["password"]

    # https://datatracker.ietf.org/doc/html/rfc7592#section-2.2
    # If omitted, the default behavior is that the client will use only
    # the "authorization_code" Grant Type.
    body = {"client_id": "client-id", "client_name": "Authlib"}
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert "grant_types" not in resp

    body = {
        "client_id": "client-id",
        "grant_types": ["client_credentials"],
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_update_token_endpoint_auth_methods_supported(test_client, metadata, token):
    metadata["token_endpoint_auth_methods_supported"] = ["client_secret_basic"]

    headers = {"Authorization": f"bearer {token.access_token}"}
    body = {
        "client_id": "client-id",
        "token_endpoint_auth_method": "client_secret_basic",
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["client_id"] == "client-id"
    assert resp["client_name"] == "Authlib"
    assert resp["token_endpoint_auth_method"] == "client_secret_basic"

    body = {
        "client_id": "client-id",
        "token_endpoint_auth_method": "none",
        "client_name": "Authlib",
    }
    rv = test_client.put("/configure_client/client-id", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_delete_client(test_client, client, token):
    assert client.client_name == "Authlib"
    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.delete("/configure_client/client-id", headers=headers)
    assert rv.status_code == 204
    assert not rv.data


def test_delete_access_denied(test_client):
    rv = test_client.delete("/configure_client/client-id")
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"

    headers = {"Authorization": "bearer invalid_token"}
    rv = test_client.delete("/configure_client/client-id", headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"

    headers = {"Authorization": "bearer unauthorized_token"}
    rv = test_client.delete(
        "/configure_client/client-id",
        json={"client_id": "client-id", "client_name": "new client_name"},
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert rv.status_code == 400
    assert resp["error"] == "access_denied"


def test_delete_invalid_client(test_client, token):
    # If the client does not exist on this server, the server MUST respond
    # with HTTP 401 Unauthorized, and the registration access token used to
    # make this request SHOULD be immediately revoked.

    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.delete("/configure_client/invalid_client_id", headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 401
    assert resp["error"] == "invalid_client"


def test_delete_unauthorized_client(test_client, token):
    # If the client does not have permission to read its record, the server
    # MUST return an HTTP 403 Forbidden.

    client = Client(
        client_id="unauthorized_client_id",
        client_secret="unauthorized_client_secret",
    )
    db.session.add(client)

    headers = {"Authorization": f"bearer {token.access_token}"}
    rv = test_client.delete("/configure_client/unauthorized_client_id", headers=headers)
    resp = json.loads(rv.data)
    assert rv.status_code == 403
    assert resp["error"] == "unauthorized_client"



import pytest
from flask import json
from flask import jsonify

from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.integrations.flask_oauth2 import current_token
from authlib.integrations.sqla_oauth2 import create_bearer_token_validator

from .models import Token
from .oauth2_server import create_bearer_header


@pytest.fixture(autouse=True)
def server(server):
    return server


@pytest.fixture(autouse=True)
def resource_server(app, db):
    require_oauth = ResourceProtector()
    BearerTokenValidator = create_bearer_token_validator(db.session, Token)
    require_oauth.register_token_validator(BearerTokenValidator())

    @app.route("/user")
    @require_oauth("profile")
    def user_profile():
        user = current_token.user
        return jsonify(id=user.id, username=user.username)

    @app.route("/user/email")
    @require_oauth("email")
    def user_email():
        user = current_token.user
        return jsonify(email=user.username + "@example.com")

    @app.route("/info")
    @require_oauth()
    def public_info():
        return jsonify(status="ok")

    @app.route("/operator-and")
    @require_oauth(["profile email"])
    def operator_and():
        return jsonify(status="ok")

    @app.route("/operator-or")
    @require_oauth(["profile", "email"])
    def operator_or():
        return jsonify(status="ok")

    @app.route("/acquire")
    def test_acquire():
        with require_oauth.acquire("profile") as token:
            user = token.user
            return jsonify(id=user.id, username=user.username)

    @app.route("/optional")
    @require_oauth("profile", optional=True)
    def test_optional_token():
        if current_token:
            user = current_token.user
            return jsonify(id=user.id, username=user.username)
        else:
            return jsonify(id=0, username="anonymous")

    return require_oauth


def test_authorization_none_grant(test_client):
    authorize_url = "/oauth/authorize?response_type=token&client_id=implicit-client"
    rv = test_client.get(authorize_url)
    assert b"unsupported_response_type" in rv.data

    rv = test_client.post(authorize_url, data={"user_id": "1"})
    assert rv.status != 200

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": "x",
        },
    )
    data = json.loads(rv.data)
    assert data["error"] == "unsupported_grant_type"


def test_invalid_token(test_client, token):
    rv = test_client.get("/user")
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "missing_authorization"

    headers = {"Authorization": "invalid token"}
    rv = test_client.get("/user", headers=headers)
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_token_type"

    headers = create_bearer_header("invalid")
    rv = test_client.get("/user", headers=headers)
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"


def test_expired_token(test_client, db, token):
    token.expires_in = -10
    db.session.add(token)
    db.session.commit()

    headers = create_bearer_header("a1")

    rv = test_client.get("/user", headers=headers)
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"

    rv = test_client.get("/acquire", headers=headers)
    assert rv.status_code == 401


def test_insufficient_token(test_client, token):
    headers = create_bearer_header("a1")
    rv = test_client.get("/user/email", headers=headers)
    assert rv.status_code == 403
    resp = json.loads(rv.data)
    assert resp["error"] == "insufficient_scope"


def test_access_resource(test_client, token):
    headers = create_bearer_header("a1")

    rv = test_client.get("/user", headers=headers)
    resp = json.loads(rv.data)
    assert resp["username"] == "foo"

    rv = test_client.get("/acquire", headers=headers)
    resp = json.loads(rv.data)
    assert resp["username"] == "foo"

    rv = test_client.get("/info", headers=headers)
    resp = json.loads(rv.data)
    assert resp["status"] == "ok"


def test_scope_operator(test_client, token):
    headers = create_bearer_header("a1")
    rv = test_client.get("/operator-and", headers=headers)
    assert rv.status_code == 403
    resp = json.loads(rv.data)
    assert resp["error"] == "insufficient_scope"

    rv = test_client.get("/operator-or", headers=headers)
    assert rv.status_code == 200


def test_optional_token(test_client, token):
    rv = test_client.get("/optional")
    assert rv.status_code == 200
    resp = json.loads(rv.data)
    assert resp["username"] == "anonymous"

    headers = create_bearer_header("a1")
    rv = test_client.get("/optional", headers=headers)
    assert rv.status_code == 200
    resp = json.loads(rv.data)
    assert resp["username"] == "foo"



import pytest

from authlib.jose.errors import InvalidClaimError
from authlib.oidc.registration import ClientMetadataClaims


def test_request_uris():
    claims = ClientMetadataClaims(
        {"request_uris": ["https://client.test/request_uris"]}, {}
    )
    claims.validate()

    claims = ClientMetadataClaims({"request_uris": ["invalid"]}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_initiate_login_uri():
    claims = ClientMetadataClaims(
        {"initiate_login_uri": "https://client.test/initiate_login_uri"}, {}
    )
    claims.validate()

    claims = ClientMetadataClaims({"initiate_login_uri": "invalid"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_token_endpoint_auth_signing_alg():
    claims = ClientMetadataClaims({"token_endpoint_auth_signing_alg": "RSA256"}, {})
    claims.validate()

    # The value none MUST NOT be used.
    claims = ClientMetadataClaims({"token_endpoint_auth_signing_alg": "none"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_id_token_signed_response_alg():
    claims = ClientMetadataClaims({"id_token_signed_response_alg": "RSA256"}, {})
    claims.validate()


def test_default_max_age():
    claims = ClientMetadataClaims({"default_max_age": 1234}, {})
    claims.validate()

    # The value none MUST NOT be used.
    claims = ClientMetadataClaims({"default_max_age": "invalid"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()



import binascii


def encode_int(num, bits):
    length = ((bits + 7) // 8) * 2
    padded_hex = f"{num:0{length}x}"
    big_endian = binascii.a2b_hex(padded_hex.encode("ascii"))
    return big_endian


def decode_int(b):
    return int(binascii.b2a_hex(b), 16)



import pytest
from flask import json

from authlib.integrations.sqla_oauth2 import create_query_token_func
from authlib.oauth2.rfc7662 import IntrospectionEndpoint

from .models import Token
from .models import User
from .models import db
from .oauth2_server import create_basic_header

query_token = create_query_token_func(db.session, Token)


class MyIntrospectionEndpoint(IntrospectionEndpoint):
    def check_permission(self, token, client, request):
        return True

    def query_token(self, token, token_type_hint):
        return query_token(token, token_type_hint)

    def introspect_token(self, token):
        user = db.session.get(User, token.user_id)
        return {
            "active": True,
            "client_id": token.client_id,
            "username": user.username,
            "scope": token.scope,
            "sub": user.get_user_id(),
            "aud": token.client_id,
            "iss": "https://provider.test/",
            "exp": token.issued_at + token.expires_in,
            "iat": token.issued_at,
        }


@pytest.fixture(autouse=True)
def server(server, app):
    server.register_endpoint(MyIntrospectionEndpoint)

    @app.route("/oauth/introspect", methods=["POST"])
    def introspect_token():
        return server.create_endpoint_response("introspection")

    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/callback"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_invalid_client(test_client):
    rv = test_client.post("/oauth/introspect")
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = {"Authorization": "invalid token_string"}
    rv = test_client.post("/oauth/introspect", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("invalid-client", "client-secret")
    rv = test_client.post("/oauth/introspect", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("client-id", "invalid-secret")
    rv = test_client.post("/oauth/introspect", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_invalid_token(test_client):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post("/oauth/introspect", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    rv = test_client.post(
        "/oauth/introspect",
        data={
            "token_type_hint": "refresh_token",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    rv = test_client.post(
        "/oauth/introspect",
        data={
            "token": "a1",
            "token_type_hint": "unsupported_token_type",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_token_type"

    rv = test_client.post(
        "/oauth/introspect",
        data={
            "token": "invalid-token",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["active"] is False

    rv = test_client.post(
        "/oauth/introspect",
        data={
            "token": "a1",
            "token_type_hint": "refresh_token",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["active"] is False


def test_introspect_token_with_hint(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/introspect",
        data={
            "token": "a1",
            "token_type_hint": "access_token",
        },
        headers=headers,
    )
    assert rv.status_code == 200
    resp = json.loads(rv.data)
    assert resp["client_id"] == "client-id"


def test_introspect_token_without_hint(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/introspect",
        data={
            "token": "a1",
        },
        headers=headers,
    )
    assert rv.status_code == 200
    resp = json.loads(rv.data)
    assert resp["client_id"] == "client-id"



"""authlib.oauth2.rfc7636.
~~~~~~~~~~~~~~~~~~~~~~

This module represents a direct implementation of
Proof Key for Code Exchange by OAuth Public Clients.

https://tools.ietf.org/html/rfc7636
"""

from .challenge import CodeChallenge
from .challenge import create_s256_code_challenge

__all__ = ["CodeChallenge", "create_s256_code_challenge"]



class BearerTokenGenerator:
    """Bearer token generator which can create the payload for token response
    by OAuth 2 server. A typical token response would be:

    .. code-block:: http

        HTTP/1.1 200 OK
        Content-Type: application/json;charset=UTF-8
        Cache-Control: no-store
        Pragma: no-cache

        {
            "access_token":"mF_9.B5f-4.1JqM",
            "token_type":"Bearer",
            "expires_in":3600,
            "refresh_token":"tGzv3JOkF0XG5Qx2TlKWIA"
        }
    """

    #: default expires_in value
    DEFAULT_EXPIRES_IN = 3600
    #: default expires_in value differentiate by grant_type
    GRANT_TYPES_EXPIRES_IN = {
        "authorization_code": 864000,
        "implicit": 3600,
        "password": 864000,
        "client_credentials": 864000,
    }

    def __init__(
        self,
        access_token_generator,
        refresh_token_generator=None,
        expires_generator=None,
    ):
        self.access_token_generator = access_token_generator
        self.refresh_token_generator = refresh_token_generator
        self.expires_generator = expires_generator

    def _get_expires_in(self, client, grant_type):
        if self.expires_generator is None:
            expires_in = self.GRANT_TYPES_EXPIRES_IN.get(
                grant_type, self.DEFAULT_EXPIRES_IN
            )
        elif callable(self.expires_generator):
            expires_in = self.expires_generator(client, grant_type)
        elif isinstance(self.expires_generator, int):
            expires_in = self.expires_generator
        else:
            expires_in = self.DEFAULT_EXPIRES_IN
        return expires_in

    @staticmethod
    def get_allowed_scope(client, scope):
        if scope:
            scope = client.get_allowed_scope(scope)
        return scope

    def generate(
        self,
        grant_type,
        client,
        user=None,
        scope=None,
        expires_in=None,
        include_refresh_token=True,
    ):
        """Generate a bearer token for OAuth 2.0 authorization token endpoint.

        :param client: the client that making the request.
        :param grant_type: current requested grant_type.
        :param user: current authorized user.
        :param expires_in: if provided, use this value as expires_in.
        :param scope: current requested scope.
        :param include_refresh_token: should refresh_token be included.
        :return: Token dict
        """
        scope = self.get_allowed_scope(client, scope)
        access_token = self.access_token_generator(
            client=client, grant_type=grant_type, user=user, scope=scope
        )
        if expires_in is None:
            expires_in = self._get_expires_in(client, grant_type)

        token = {
            "token_type": "Bearer",
            "access_token": access_token,
        }
        if expires_in:
            token["expires_in"] = expires_in
        if include_refresh_token and self.refresh_token_generator:
            token["refresh_token"] = self.refresh_token_generator(
                client=client, grant_type=grant_type, user=user, scope=scope
            )
        if scope:
            token["scope"] = scope
        return token

    def __call__(
        self,
        grant_type,
        client,
        user=None,
        scope=None,
        expires_in=None,
        include_refresh_token=True,
    ):
        return self.generate(
            grant_type, client, user, scope, expires_in, include_refresh_token
        )



"""
authlib
~~~~~~~

The ultimate Python library in building OAuth 1.0, OAuth 2.0 and OpenID
Connect clients and providers. It covers from low level specification
implementation to high level framework integrations.

:copyright: (c) 2017 by Hsiaoming Yang.
:license: BSD, see LICENSE for more details.
"""

from .consts import author
from .consts import homepage
from .consts import version

__version__ = version
__homepage__ = homepage
__author__ = author
__license__ = "BSD-3-Clause"



from authlib.common.urls import add_params_to_qs

from .assertion import client_secret_jwt_sign
from .assertion import private_key_jwt_sign
from .client import ASSERTION_TYPE


class ClientSecretJWT:
    """Authentication method for OAuth 2.0 Client. This authentication
    method is called ``client_secret_jwt``, which is using ``client_id``
    and ``client_secret`` constructed with JWT to identify a client.

    Here is an example of use ``client_secret_jwt`` with Requests Session::

        from authlib.integrations.requests_client import OAuth2Session

        token_endpoint = "https://example.com/oauth/token"
        session = OAuth2Session(
            "your-client-id",
            "your-client-secret",
            token_endpoint_auth_method="client_secret_jwt",
        )
        session.register_client_auth_method(ClientSecretJWT(token_endpoint))
        session.fetch_token(token_endpoint)

    :param token_endpoint: A string URL of the token endpoint
    :param claims: Extra JWT claims
    :param headers: Extra JWT headers
    :param alg: ``alg`` value, default is HS256
    """

    name = "client_secret_jwt"
    alg = "HS256"

    def __init__(self, token_endpoint=None, claims=None, headers=None, alg=None):
        self.token_endpoint = token_endpoint
        self.claims = claims
        self.headers = headers
        if alg is not None:
            self.alg = alg

    def sign(self, auth, token_endpoint):
        return client_secret_jwt_sign(
            auth.client_secret,
            client_id=auth.client_id,
            token_endpoint=token_endpoint,
            claims=self.claims,
            header=self.headers,
            alg=self.alg,
        )

    def __call__(self, auth, method, uri, headers, body):
        token_endpoint = self.token_endpoint
        if not token_endpoint:
            token_endpoint = uri

        client_assertion = self.sign(auth, token_endpoint)
        body = add_params_to_qs(
            body or "",
            [
                ("client_assertion_type", ASSERTION_TYPE),
                ("client_assertion", client_assertion),
            ],
        )
        return uri, headers, body


class PrivateKeyJWT(ClientSecretJWT):
    """Authentication method for OAuth 2.0 Client. This authentication
    method is called ``private_key_jwt``, which is using ``client_id``
    and ``private_key`` constructed with JWT to identify a client.

    Here is an example of use ``private_key_jwt`` with Requests Session::

        from authlib.integrations.requests_client import OAuth2Session

        token_endpoint = "https://example.com/oauth/token"
        session = OAuth2Session(
            "your-client-id",
            "your-client-private-key",
            token_endpoint_auth_method="private_key_jwt",
        )
        session.register_client_auth_method(PrivateKeyJWT(token_endpoint))
        session.fetch_token(token_endpoint)

    :param token_endpoint: A string URL of the token endpoint
    :param claims: Extra JWT claims
    :param headers: Extra JWT headers
    :param alg: ``alg`` value, default is RS256
    """

    name = "private_key_jwt"
    alg = "RS256"

    def sign(self, auth, token_endpoint):
        return private_key_jwt_sign(
            auth.client_secret,
            client_id=auth.client_id,
            token_endpoint=token_endpoint,
            claims=self.claims,
            header=self.headers,
            alg=self.alg,
        )



import json

from starlette.requests import Request as ASGIRequest
from starlette.responses import Response as ASGIResponse


class AsyncMockDispatch:
    def __init__(self, body=b"", status_code=200, headers=None, assert_func=None):
        if headers is None:
            headers = {}
        if isinstance(body, dict):
            body = json.dumps(body).encode()
            headers["Content-Type"] = "application/json"
        else:
            if isinstance(body, str):
                body = body.encode()
            headers["Content-Type"] = "application/x-www-form-urlencoded"

        self.body = body
        self.status_code = status_code
        self.headers = headers
        self.assert_func = assert_func

    async def __call__(self, scope, receive, send):
        request = ASGIRequest(scope, receive=receive)

        if self.assert_func:
            await self.assert_func(request)

        response = ASGIResponse(
            status_code=self.status_code,
            content=self.body,
            headers=self.headers,
        )
        await response(scope, receive, send)


class AsyncPathMapDispatch:
    def __init__(self, path_maps):
        self.path_maps = path_maps

    async def __call__(self, scope, receive, send):
        request = ASGIRequest(scope, receive=receive)

        rv = self.path_maps[request.url.path]
        status_code = rv.get("status_code", 200)
        body = rv.get("body")
        headers = rv.get("headers", {})
        if isinstance(body, dict):
            body = json.dumps(body).encode()
            headers["Content-Type"] = "application/json"
        else:
            if isinstance(body, str):
                body = body.encode()
            headers["Content-Type"] = "application/x-www-form-urlencoded"

        response = ASGIResponse(
            status_code=status_code,
            content=body,
            headers=headers,
        )
        await response(scope, receive, send)



import pytest
from httpx import ASGITransport
from starlette.requests import Request

from authlib.integrations.starlette_client import OAuth
from authlib.jose import JsonWebKey
from authlib.jose.errors import InvalidClaimError
from authlib.oidc.core.grants.util import generate_id_token

from ..asgi_helper import AsyncPathMapDispatch
from ..util import get_bearer_token
from ..util import read_key_file

secret_key = JsonWebKey.import_key("secret", {"kty": "oct", "kid": "f"})


async def run_fetch_userinfo(payload):
    oauth = OAuth()

    async def fetch_token(request):
        return get_bearer_token()

    transport = ASGITransport(AsyncPathMapDispatch({"/userinfo": {"body": payload}}))

    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=fetch_token,
        userinfo_endpoint="https://provider.test/userinfo",
        client_kwargs={
            "transport": transport,
        },
    )

    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    user = await client.userinfo(request=req)
    assert user.sub == "123"


@pytest.mark.asyncio
async def test_fetch_userinfo():
    await run_fetch_userinfo({"sub": "123"})


@pytest.mark.asyncio
async def test_parse_id_token():
    token = get_bearer_token()
    id_token = generate_id_token(
        token,
        {"sub": "123"},
        secret_key,
        alg="HS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce="n",
    )
    token["id_token"] = id_token

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        jwks={"keys": [secret_key.as_dict()]},
        issuer="https://provider.test",
        id_token_signing_alg_values_supported=["HS256", "RS256"],
    )
    user = await client.parse_id_token(token, nonce="n")
    assert user.sub == "123"

    claims_options = {"iss": {"value": "https://provider.test"}}
    user = await client.parse_id_token(token, nonce="n", claims_options=claims_options)
    assert user.sub == "123"

    with pytest.raises(InvalidClaimError):
        claims_options = {"iss": {"value": "https://wrong-provider.test"}}
        await client.parse_id_token(token, nonce="n", claims_options=claims_options)


@pytest.mark.asyncio
async def test_runtime_error_fetch_jwks_uri():
    token = get_bearer_token()
    id_token = generate_id_token(
        token,
        {"sub": "123"},
        secret_key,
        alg="HS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce="n",
    )

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        issuer="https://provider.test",
        id_token_signing_alg_values_supported=["HS256"],
    )
    req_scope = {"type": "http", "session": {"_dev_authlib_nonce_": "n"}}
    req = Request(req_scope)
    token["id_token"] = id_token
    with pytest.raises(RuntimeError):
        await client.parse_id_token(req, token)


@pytest.mark.asyncio
async def test_force_fetch_jwks_uri():
    secret_keys = read_key_file("jwks_private.json")
    token = get_bearer_token()
    id_token = generate_id_token(
        token,
        {"sub": "123"},
        secret_keys,
        alg="RS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce="n",
    )
    token["id_token"] = id_token

    transport = ASGITransport(
        AsyncPathMapDispatch({"/jwks": {"body": read_key_file("jwks_public.json")}})
    )

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        jwks_uri="https://provider.test/jwks",
        issuer="https://provider.test",
        client_kwargs={
            "transport": transport,
        },
    )
    user = await client.parse_id_token(token, nonce="n")
    assert user.sub == "123"



import time

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import Text

from authlib.oauth2.rfc6749 import AuthorizationCodeMixin
from authlib.oauth2.rfc6749 import TokenMixin


class OAuth2AuthorizationCodeMixin(AuthorizationCodeMixin):
    code = Column(String(120), unique=True, nullable=False)
    client_id = Column(String(48))
    redirect_uri = Column(Text, default="")
    response_type = Column(Text, default="")
    scope = Column(Text, default="")
    nonce = Column(Text)
    auth_time = Column(Integer, nullable=False, default=lambda: int(time.time()))
    acr = Column(Text, nullable=True)
    amr = Column(Text, nullable=True)

    code_challenge = Column(Text)
    code_challenge_method = Column(String(48))

    def is_expired(self):
        return self.auth_time + 300 < time.time()

    def get_redirect_uri(self):
        return self.redirect_uri

    def get_scope(self):
        return self.scope

    def get_auth_time(self):
        return self.auth_time

    def get_acr(self):
        return self.acr

    def get_amr(self):
        return self.amr.split() if self.amr else []

    def get_nonce(self):
        return self.nonce


class OAuth2TokenMixin(TokenMixin):
    client_id = Column(String(48))
    token_type = Column(String(40))
    access_token = Column(String(255), unique=True, nullable=False)
    refresh_token = Column(String(255), index=True)
    scope = Column(Text, default="")
    issued_at = Column(Integer, nullable=False, default=lambda: int(time.time()))
    access_token_revoked_at = Column(Integer, nullable=False, default=0)
    refresh_token_revoked_at = Column(Integer, nullable=False, default=0)
    expires_in = Column(Integer, nullable=False, default=0)

    def check_client(self, client):
        return self.client_id == client.get_client_id()

    def get_scope(self):
        return self.scope

    def get_expires_in(self):
        return self.expires_in

    def is_revoked(self):
        return self.access_token_revoked_at or self.refresh_token_revoked_at

    def is_expired(self):
        if not self.expires_in:
            return False

        expires_at = self.issued_at + self.expires_in
        return expires_at < time.time()



import pytest

from authlib.jose.errors import InvalidClaimError
from authlib.oauth2.rfc7591 import ClientMetadataClaims


def test_validate_redirect_uris():
    claims = ClientMetadataClaims({"redirect_uris": ["foo"]}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_validate_client_uri():
    claims = ClientMetadataClaims({"client_uri": "foo"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_validate_logo_uri():
    claims = ClientMetadataClaims({"logo_uri": "foo"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_validate_tos_uri():
    claims = ClientMetadataClaims({"tos_uri": "foo"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_validate_policy_uri():
    claims = ClientMetadataClaims({"policy_uri": "foo"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()


def test_validate_jwks_uri():
    claims = ClientMetadataClaims({"jwks_uri": "foo"}, {})
    with pytest.raises(InvalidClaimError):
        claims.validate()



import json
import time


class FrameworkIntegration:
    expires_in = 3600

    def __init__(self, name, cache=None):
        self.name = name
        self.cache = cache

    def _get_cache_data(self, key):
        value = self.cache.get(key)
        if not value:
            return None
        try:
            return json.loads(value)
        except (TypeError, ValueError):
            return None

    def _clear_session_state(self, session):
        now = time.time()
        prefix = f"_state_{self.name}"
        for key in dict(session):
            if key.startswith(prefix):
                value = session[key]
                exp = value.get("exp")
                if not exp or exp < now:
                    session.pop(key)

    def get_state_data(self, session, state):
        key = f"_state_{self.name}_{state}"
        session_data = session.get(key)
        if not session_data:
            return None
        if self.cache:
            cached_value = self._get_cache_data(key)
        else:
            cached_value = session_data
        if cached_value:
            return cached_value.get("data")
        return None

    def set_state_data(self, session, state, data):
        key = f"_state_{self.name}_{state}"
        now = time.time()
        if self.cache:
            self.cache.set(key, json.dumps({"data": data}), self.expires_in)
            session[key] = {"exp": now + self.expires_in}
        else:
            session[key] = {"data": data, "exp": now + self.expires_in}

    def clear_state_data(self, session, state):
        key = f"_state_{self.name}_{state}"
        if self.cache:
            self.cache.delete(key)
        session.pop(key, None)
        self._clear_session_state(session)

    def update_token(self, token, refresh_token=None, access_token=None):
        raise NotImplementedError()

    @staticmethod
    def load_config(oauth, name, params):
        raise NotImplementedError()



from authlib.consts import default_json_headers

from ..errors import InvalidRequestError
from ..hooks import Hookable
from ..hooks import hooked
from ..requests import OAuth2Request


class BaseGrant(Hookable):
    #: Allowed client auth methods for token endpoint
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic"]

    #: Designed for which "grant_type"
    GRANT_TYPE = None

    # NOTE: there is no charset for application/json, since
    # application/json should always in UTF-8.
    # The example on RFC is incorrect.
    # https://tools.ietf.org/html/rfc4627
    TOKEN_RESPONSE_HEADER = default_json_headers

    def __init__(self, request: OAuth2Request, server):
        super().__init__()
        self.prompt = None
        self.redirect_uri = None
        self.request = request
        self.server = server

    @property
    def client(self):
        return self.request.client

    def generate_token(
        self,
        user=None,
        scope=None,
        grant_type=None,
        expires_in=None,
        include_refresh_token=True,
    ):
        if grant_type is None:
            grant_type = self.GRANT_TYPE
        return self.server.generate_token(
            client=self.request.client,
            grant_type=grant_type,
            user=user,
            scope=scope,
            expires_in=expires_in,
            include_refresh_token=include_refresh_token,
        )

    def authenticate_token_endpoint_client(self):
        """Authenticate client with the given methods for token endpoint.

        For example, the client makes the following HTTP request using TLS:

        .. code-block:: http

            POST /token HTTP/1.1
            Host: server.example.com
            Authorization: Basic czZCaGRSa3F0MzpnWDFmQmF0M2JW
            Content-Type: application/x-www-form-urlencoded

            grant_type=authorization_code&code=SplxlOBeZQQYbYS6WxSbIA
            &redirect_uri=https%3A%2F%2Fclient%2Eexample%2Ecom%2Fcb

        Default available methods are: "none", "client_secret_basic" and
        "client_secret_post".

        :return: client
        """
        client = self.server.authenticate_client(
            self.request, self.TOKEN_ENDPOINT_AUTH_METHODS
        )
        self.server.send_signal("after_authenticate_client", client=client, grant=self)
        return client

    def save_token(self, token):
        """A method to save token into database."""
        return self.server.save_token(token, self.request)

    def validate_requested_scope(self):
        """Validate if requested scope is supported by Authorization Server."""
        scope = self.request.payload.scope
        return self.server.validate_requested_scope(scope)


class TokenEndpointMixin:
    #: Allowed HTTP methods of this token endpoint
    TOKEN_ENDPOINT_HTTP_METHODS = ["POST"]

    #: Designed for which "grant_type"
    GRANT_TYPE = None

    @classmethod
    def check_token_endpoint(cls, request: OAuth2Request):
        return (
            request.payload.grant_type == cls.GRANT_TYPE
            and request.method in cls.TOKEN_ENDPOINT_HTTP_METHODS
        )

    def validate_token_request(self):
        raise NotImplementedError()

    def create_token_response(self):
        raise NotImplementedError()


class AuthorizationEndpointMixin:
    RESPONSE_TYPES = set()
    ERROR_RESPONSE_FRAGMENT = False

    @classmethod
    def check_authorization_endpoint(cls, request: OAuth2Request):
        return request.payload.response_type in cls.RESPONSE_TYPES

    @staticmethod
    def validate_authorization_redirect_uri(request: OAuth2Request, client):
        if request.payload.redirect_uri:
            if not client.check_redirect_uri(request.payload.redirect_uri):
                raise InvalidRequestError(
                    f"Redirect URI {request.payload.redirect_uri} is not supported by client.",
                )
            return request.payload.redirect_uri
        else:
            redirect_uri = client.get_default_redirect_uri()
            if not redirect_uri:
                raise InvalidRequestError(
                    "Missing 'redirect_uri' in request.", state=request.payload.state
                )
            return redirect_uri

    @staticmethod
    def validate_no_multiple_request_parameter(request: OAuth2Request):
        """For the Authorization Endpoint, request and response parameters MUST NOT be included
        more than once. Per `Section 3.1`_.

        .. _`Section 3.1`: https://tools.ietf.org/html/rfc6749#section-3.1
        """
        datalist = request.payload.datalist
        parameters = ["response_type", "client_id", "redirect_uri", "scope", "state"]
        for param in parameters:
            if len(datalist.get(param, [])) > 1:
                raise InvalidRequestError(
                    f"Multiple '{param}' in request.", state=request.payload.state
                )

    @hooked
    def validate_consent_request(self):
        redirect_uri = self.validate_authorization_request()
        self.redirect_uri = redirect_uri
        return redirect_uri

    def validate_authorization_request(self):
        raise NotImplementedError()

    def create_authorization_response(self, redirect_uri: str, grant_user):
        raise NotImplementedError()



import logging
import time

from authlib.jose import JoseError
from authlib.jose import JWTClaims
from authlib.jose import jwt

from ..rfc6749 import TokenMixin
from ..rfc6750 import BearerTokenValidator

logger = logging.getLogger(__name__)


class JWTBearerToken(TokenMixin, JWTClaims):
    def check_client(self, client):
        return self["client_id"] == client.get_client_id()

    def get_scope(self):
        return self.get("scope")

    def get_expires_in(self):
        return self["exp"] - self["iat"]

    def is_expired(self):
        return self["exp"] < time.time()

    def is_revoked(self):
        return False


class JWTBearerTokenValidator(BearerTokenValidator):
    TOKEN_TYPE = "bearer"
    token_cls = JWTBearerToken

    def __init__(self, public_key, issuer=None, realm=None, **extra_attributes):
        super().__init__(realm, **extra_attributes)
        self.public_key = public_key
        claims_options = {
            "exp": {"essential": True},
            "client_id": {"essential": True},
            "grant_type": {"essential": True},
        }
        if issuer:
            claims_options["iss"] = {"essential": True, "value": issuer}
        self.claims_options = claims_options

    def authenticate_token(self, token_string):
        try:
            claims = jwt.decode(
                token_string,
                self.public_key,
                claims_options=self.claims_options,
                claims_cls=self.token_cls,
            )
            claims.validate()
            return claims
        except JoseError as error:
            logger.debug("Authenticate token failed. %r", error)
            return None



import pytest

from authlib.jose.rfc7518.ec_key import ECKey
from authlib.jose.rfc7518.oct_key import OctKey
from authlib.jose.rfc7518.rsa_key import RSAKey
from authlib.jose.rfc8037.okp_key import OKPKey
from authlib.oidc.core import UserInfo
from authlib.oidc.core.grants.util import generate_id_token

hmac_key = OctKey.generate_key(256)
rsa_key = RSAKey.generate_key(2048, is_private=True)
ec_key = ECKey.generate_key("P-256", is_private=True)
okp_key = OKPKey.generate_key("Ed25519", is_private=True)
ec_secp256k1_key = ECKey.generate_key("secp256k1", is_private=True)


@pytest.mark.parametrize(
    "alg,key",
    [
        ("none", None),
        ("HS256", hmac_key),
        ("HS384", hmac_key),
        ("HS512", hmac_key),
        ("RS256", rsa_key),
        ("RS384", rsa_key),
        ("RS512", rsa_key),
        ("ES256", ec_key),
        ("PS256", rsa_key),
        ("PS384", rsa_key),
        ("PS512", rsa_key),
        ("EdDSA", okp_key),
        ("ES256K", ec_secp256k1_key),
    ],
)
def test_generate_id_token(alg, key):
    token = {"access_token": "test_token"}
    user_info = UserInfo({"sub": "123"})

    result = generate_id_token(
        token=token,
        user_info=user_info,
        key=key,
        iss="https://provider.test",
        aud="client_id",
        alg=alg,
    )
    assert result is not None



"""authlib.oauth2.rfc8628.
~~~~~~~~~~~~~~~~~~~~~~

This module represents an implementation of
OAuth 2.0 Device Authorization Grant.

https://tools.ietf.org/html/rfc8628
"""

from .device_code import DEVICE_CODE_GRANT_TYPE
from .device_code import DeviceCodeGrant
from .endpoint import DeviceAuthorizationEndpoint
from .errors import AuthorizationPendingError
from .errors import ExpiredTokenError
from .errors import SlowDownError
from .models import DeviceCredentialDict
from .models import DeviceCredentialMixin

__all__ = [
    "DeviceAuthorizationEndpoint",
    "DeviceCodeGrant",
    "DEVICE_CODE_GRANT_TYPE",
    "DeviceCredentialMixin",
    "DeviceCredentialDict",
    "AuthorizationPendingError",
    "SlowDownError",
    "ExpiredTokenError",
]



"""authlib.jose
~~~~~~~~~~~~

JOSE implementation in Authlib. Tracking the status of JOSE specs at
https://tools.ietf.org/wg/jose/
"""

from .errors import JoseError
from .rfc7515 import JsonWebSignature
from .rfc7515 import JWSAlgorithm
from .rfc7515 import JWSHeader
from .rfc7515 import JWSObject
from .rfc7516 import JsonWebEncryption
from .rfc7516 import JWEAlgorithm
from .rfc7516 import JWEEncAlgorithm
from .rfc7516 import JWEZipAlgorithm
from .rfc7517 import JsonWebKey
from .rfc7517 import Key
from .rfc7517 import KeySet
from .rfc7518 import ECDHESAlgorithm
from .rfc7518 import ECKey
from .rfc7518 import OctKey
from .rfc7518 import RSAKey
from .rfc7518 import register_jwe_rfc7518
from .rfc7518 import register_jws_rfc7518
from .rfc7519 import BaseClaims
from .rfc7519 import JsonWebToken
from .rfc7519 import JWTClaims
from .rfc8037 import OKPKey
from .rfc8037 import register_jws_rfc8037

# register algorithms
register_jws_rfc7518(JsonWebSignature)
register_jws_rfc8037(JsonWebSignature)

register_jwe_rfc7518(JsonWebEncryption)

# attach algorithms
ECDHESAlgorithm.ALLOWED_KEY_CLS = (ECKey, OKPKey)

# register supported keys
JsonWebKey.JWK_KEY_CLS = {
    OctKey.kty: OctKey,
    RSAKey.kty: RSAKey,
    ECKey.kty: ECKey,
    OKPKey.kty: OKPKey,
}

jwt = JsonWebToken(
    [
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "PS256",
        "PS384",
        "PS512",
    ]
)


__all__ = [
    "JoseError",
    "JsonWebSignature",
    "JWSAlgorithm",
    "JWSHeader",
    "JWSObject",
    "JsonWebEncryption",
    "JWEAlgorithm",
    "JWEEncAlgorithm",
    "JWEZipAlgorithm",
    "JsonWebKey",
    "Key",
    "KeySet",
    "OctKey",
    "RSAKey",
    "ECKey",
    "OKPKey",
    "JsonWebToken",
    "BaseClaims",
    "JWTClaims",
    "jwt",
]



from unittest import mock

import pytest
from flask import Flask

from authlib.integrations.flask_client import OAuth
from authlib.jose import JsonWebKey
from authlib.jose.errors import InvalidClaimError
from authlib.oidc.core.grants.util import generate_id_token

from ..util import get_bearer_token
from ..util import read_key_file

secret_key = JsonWebKey.import_key("secret", {"kty": "oct", "kid": "f"})


def test_fetch_userinfo():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        userinfo_endpoint="https://provider.test/userinfo",
    )

    def fake_send(sess, req, **kwargs):
        resp = mock.MagicMock()
        resp.json = lambda: {"sub": "123"}
        resp.status_code = 200
        return resp

    with app.test_request_context():
        with mock.patch("requests.sessions.Session.send", fake_send):
            user = client.userinfo()
            assert user.sub == "123"


def test_parse_id_token():
    token = get_bearer_token()
    id_token = generate_id_token(
        token,
        {"sub": "123"},
        secret_key,
        alg="HS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce="n",
    )

    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        jwks={"keys": [secret_key.as_dict()]},
        issuer="https://provider.test",
        id_token_signing_alg_values_supported=["HS256", "RS256"],
    )
    with app.test_request_context():
        assert client.parse_id_token(token, nonce="n") is None

        token["id_token"] = id_token
        user = client.parse_id_token(token, nonce="n")
        assert user.sub == "123"

        claims_options = {"iss": {"value": "https://provider.test"}}
        user = client.parse_id_token(token, nonce="n", claims_options=claims_options)
        assert user.sub == "123"

        claims_options = {"iss": {"value": "https://wrong-provider.test"}}
        with pytest.raises(InvalidClaimError):
            client.parse_id_token(token, "n", claims_options)


def test_parse_id_token_nonce_supported():
    token = get_bearer_token()
    id_token = generate_id_token(
        token,
        {"sub": "123", "nonce_supported": False},
        secret_key,
        alg="HS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
    )

    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        jwks={"keys": [secret_key.as_dict()]},
        issuer="https://provider.test",
        id_token_signing_alg_values_supported=["HS256", "RS256"],
    )
    with app.test_request_context():
        token["id_token"] = id_token
        user = client.parse_id_token(token, nonce="n")
        assert user.sub == "123"


def test_runtime_error_fetch_jwks_uri():
    token = get_bearer_token()
    id_token = generate_id_token(
        token,
        {"sub": "123"},
        secret_key,
        alg="HS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce="n",
    )

    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    alt_key = secret_key.as_dict()
    alt_key["kid"] = "b"
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        jwks={"keys": [alt_key]},
        issuer="https://provider.test",
        id_token_signing_alg_values_supported=["HS256"],
    )
    with app.test_request_context():
        token["id_token"] = id_token
        with pytest.raises(RuntimeError):
            client.parse_id_token(token, "n")


def test_force_fetch_jwks_uri():
    secret_keys = read_key_file("jwks_private.json")
    token = get_bearer_token()
    id_token = generate_id_token(
        token,
        {"sub": "123"},
        secret_keys,
        alg="RS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce="n",
    )

    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        fetch_token=get_bearer_token,
        jwks={"keys": [secret_key.as_dict()]},
        jwks_uri="https://provider.test/jwks",
        issuer="https://provider.test",
    )

    def fake_send(sess, req, **kwargs):
        resp = mock.MagicMock()
        resp.json = lambda: read_key_file("jwks_public.json")
        resp.status_code = 200
        return resp

    with app.test_request_context():
        assert client.parse_id_token(token, nonce="n") is None

        with mock.patch("requests.sessions.Session.send", fake_send):
            token["id_token"] = id_token
            user = client.parse_id_token(token, nonce="n")
            assert user.sub == "123"



from authlib.jose import JsonWebSignature
from tests.util import read_file_path


def test_EdDSA_alg():
    jws = JsonWebSignature(algorithms=["EdDSA"])
    private_key = read_file_path("ed25519-pkcs8.pem")
    public_key = read_file_path("ed25519-pub.pem")
    s = jws.serialize({"alg": "EdDSA"}, "hello", private_key)
    data = jws.deserialize(s, public_key)
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "EdDSA"



import pytest
from httpx import ASGITransport

from authlib.integrations.httpx_client import SIGNATURE_TYPE_BODY
from authlib.integrations.httpx_client import SIGNATURE_TYPE_QUERY
from authlib.integrations.httpx_client import AsyncOAuth1Client
from authlib.integrations.httpx_client import OAuthError

from ..asgi_helper import AsyncMockDispatch

oauth_url = "https://provider.test/oauth"


@pytest.mark.asyncio
async def test_fetch_request_token_via_header():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    async def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert 'oauth_consumer_key="id"' in auth_header
        assert "oauth_signature=" in auth_header

    transport = ASGITransport(AsyncMockDispatch(request_token, assert_func=assert_func))
    async with AsyncOAuth1Client("id", "secret", transport=transport) as client:
        response = await client.fetch_request_token(oauth_url)

    assert response == request_token


@pytest.mark.asyncio
async def test_fetch_request_token_via_body():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    async def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert auth_header is None

        content = await request.body()
        assert b"oauth_consumer_key=id" in content
        assert b"&oauth_signature=" in content

    transport = ASGITransport(AsyncMockDispatch(request_token, assert_func=assert_func))

    async with AsyncOAuth1Client(
        "id",
        "secret",
        signature_type=SIGNATURE_TYPE_BODY,
        transport=transport,
    ) as client:
        response = await client.fetch_request_token(oauth_url)

    assert response == request_token


@pytest.mark.asyncio
async def test_fetch_request_token_via_query():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    async def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert auth_header is None

        url = str(request.url)
        assert "oauth_consumer_key=id" in url
        assert "&oauth_signature=" in url

    transport = ASGITransport(AsyncMockDispatch(request_token, assert_func=assert_func))

    async with AsyncOAuth1Client(
        "id",
        "secret",
        signature_type=SIGNATURE_TYPE_QUERY,
        transport=transport,
    ) as client:
        response = await client.fetch_request_token(oauth_url)

    assert response == request_token


@pytest.mark.asyncio
async def test_fetch_access_token():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    async def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert 'oauth_verifier="d"' in auth_header
        assert 'oauth_token="foo"' in auth_header
        assert 'oauth_consumer_key="id"' in auth_header
        assert "oauth_signature=" in auth_header

    transport = ASGITransport(AsyncMockDispatch(request_token, assert_func=assert_func))
    async with AsyncOAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        transport=transport,
    ) as client:
        with pytest.raises(OAuthError):
            await client.fetch_access_token(oauth_url)

        response = await client.fetch_access_token(oauth_url, verifier="d")

    assert response == request_token


@pytest.mark.asyncio
async def test_get_via_header():
    transport = ASGITransport(AsyncMockDispatch(b"hello"))
    async with AsyncOAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        transport=transport,
    ) as client:
        response = await client.get("https://resource.test/")

    assert response.content == b"hello"
    request = response.request
    auth_header = request.headers.get("authorization")
    assert 'oauth_token="foo"' in auth_header
    assert 'oauth_consumer_key="id"' in auth_header
    assert "oauth_signature=" in auth_header


@pytest.mark.asyncio
async def test_get_via_body():
    async def assert_func(request):
        content = await request.body()
        assert b"oauth_token=foo" in content
        assert b"oauth_consumer_key=id" in content
        assert b"oauth_signature=" in content

    transport = ASGITransport(AsyncMockDispatch(b"hello", assert_func=assert_func))
    async with AsyncOAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        signature_type=SIGNATURE_TYPE_BODY,
        transport=transport,
    ) as client:
        response = await client.post("https://resource.test/")

    assert response.content == b"hello"

    request = response.request
    auth_header = request.headers.get("authorization")
    assert auth_header is None


@pytest.mark.asyncio
async def test_get_via_query():
    transport = ASGITransport(AsyncMockDispatch(b"hello"))
    async with AsyncOAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        signature_type=SIGNATURE_TYPE_QUERY,
        transport=transport,
    ) as client:
        response = await client.get("https://resource.test/")

    assert response.content == b"hello"
    request = response.request
    auth_header = request.headers.get("authorization")
    assert auth_header is None

    url = str(request.url)
    assert "oauth_token=foo" in url
    assert "oauth_consumer_key=id" in url
    assert "oauth_signature=" in url



import pytest
from flask import json

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.jose import jwt
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oidc.core import HybridIDToken
from authlib.oidc.core.grants import OpenIDCode as _OpenIDCode
from authlib.oidc.core.grants import OpenIDHybridGrant as _OpenIDHybridGrant

from .models import CodeGrantMixin
from .models import exists_nonce
from .models import save_authorization_code
from .oauth2_server import create_basic_header

JWT_CONFIG = {"iss": "Authlib", "key": "secret", "alg": "HS256", "exp": 3600}


@pytest.fixture(autouse=True)
def server(server):
    class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
        def save_authorization_code(self, code, request):
            return save_authorization_code(code, request)

    class OpenIDCode(_OpenIDCode):
        def get_jwt_config(self, grant, client):
            return dict(JWT_CONFIG)

        def exists_nonce(self, nonce, request):
            return exists_nonce(nonce, request)

        def generate_user_info(self, user, scopes):
            return user.generate_user_info(scopes)

    class OpenIDHybridGrant(_OpenIDHybridGrant):
        def save_authorization_code(self, code, request):
            return save_authorization_code(code, request)

        def get_jwt_config(self, client):
            return dict(JWT_CONFIG)

        def exists_nonce(self, nonce, request):
            return exists_nonce(nonce, request)

        def generate_user_info(self, user, scopes):
            return user.generate_user_info(scopes)

    server.register_grant(OpenIDHybridGrant)
    server.register_grant(AuthorizationCodeGrant, [OpenIDCode()])

    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "openid profile address",
            "response_types": [
                "code id_token",
                "code token",
                "code id_token token",
            ],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def validate_claims(id_token, params):
    claims = jwt.decode(
        id_token, "secret", claims_cls=HybridIDToken, claims_params=params
    )
    claims.validate()


def test_invalid_client_id(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code token",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "invalid-client",
            "response_type": "code token",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_require_nonce(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code token",
            "scope": "openid profile",
            "state": "bar",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "error=invalid_request" in rv.location
    assert "nonce" in rv.location


def test_invalid_response_type(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code id_token invalid",
            "state": "bar",
            "nonce": "abc",
            "scope": "profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["error"] == "unsupported_response_type"


def test_invalid_scope(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code id_token",
            "state": "bar",
            "nonce": "abc",
            "scope": "profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "error=invalid_scope" in rv.location


def test_access_denied(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code token",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
        },
    )
    assert "error=access_denied" in rv.location


def test_code_access_token(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code token",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "code=" in rv.location
    assert "access_token=" in rv.location
    assert "id_token=" not in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
    assert params["state"] == "bar"

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "id_token" in resp


def test_code_id_token(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code id_token",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "code=" in rv.location
    assert "id_token=" in rv.location
    assert "access_token=" not in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
    assert params["state"] == "bar"

    params["nonce"] = "abc"
    params["client_id"] = "client-id"
    validate_claims(params["id_token"], params)

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "id_token" in resp


def test_code_id_token_access_token(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code id_token token",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "code=" in rv.location
    assert "id_token=" in rv.location
    assert "access_token=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
    assert params["state"] == "bar"
    validate_claims(params["id_token"], params)

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "id_token" in resp


def test_response_mode_query(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code id_token token",
            "response_mode": "query",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "code=" in rv.location
    assert "id_token=" in rv.location
    assert "access_token=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["state"] == "bar"


def test_response_mode_form_post(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "client_id": "client-id",
            "response_type": "code id_token token",
            "response_mode": "form_post",
            "state": "bar",
            "nonce": "abc",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert b'name="code"' in rv.data
    assert b'name="id_token"' in rv.data
    assert b'name="access_token"' in rv.data



import functools

from flask import Response
from flask import g
from flask import json
from flask import request as _req
from werkzeug.local import LocalProxy

from authlib.consts import default_json_headers
from authlib.oauth1 import ResourceProtector as _ResourceProtector
from authlib.oauth1.errors import OAuth1Error


class ResourceProtector(_ResourceProtector):
    """A protecting method for resource servers. Initialize a resource
    protector with the these method:

    1. query_client
    2. query_token,
    3. exists_nonce

    Usually, a ``query_client`` method would look like (if using SQLAlchemy)::

        def query_client(client_id):
            return Client.query.filter_by(client_id=client_id).first()

    A ``query_token`` method accept two parameters, ``client_id`` and ``oauth_token``::

        def query_token(client_id, oauth_token):
            return Token.query.filter_by(
                client_id=client_id, oauth_token=oauth_token
            ).first()

    And for ``exists_nonce``, if using cache, we have a built-in hook to create this method::

        from authlib.integrations.flask_oauth1 import create_exists_nonce_func

        exists_nonce = create_exists_nonce_func(cache)

    Then initialize the resource protector with those methods::

        require_oauth = ResourceProtector(
            app,
            query_client=query_client,
            query_token=query_token,
            exists_nonce=exists_nonce,
        )
    """

    def __init__(
        self, app=None, query_client=None, query_token=None, exists_nonce=None
    ):
        self.query_client = query_client
        self.query_token = query_token
        self._exists_nonce = exists_nonce

        self.app = app
        if app:
            self.init_app(app)

    def init_app(self, app, query_client=None, query_token=None, exists_nonce=None):
        if query_client is not None:
            self.query_client = query_client
        if query_token is not None:
            self.query_token = query_token
        if exists_nonce is not None:
            self._exists_nonce = exists_nonce

        methods = app.config.get("OAUTH1_SUPPORTED_SIGNATURE_METHODS")
        if methods and isinstance(methods, (list, tuple)):
            self.SUPPORTED_SIGNATURE_METHODS = methods

        self.app = app

    def get_client_by_id(self, client_id):
        return self.query_client(client_id)

    def get_token_credential(self, request):
        return self.query_token(request.client_id, request.token)

    def exists_nonce(self, nonce, request):
        if not self._exists_nonce:
            raise RuntimeError('"exists_nonce" function is required.')

        timestamp = request.timestamp
        client_id = request.client_id
        token = request.token
        return self._exists_nonce(nonce, timestamp, client_id, token)

    def acquire_credential(self):
        req = self.validate_request(
            _req.method, _req.url, _req.form.to_dict(flat=True), _req.headers
        )
        g.authlib_server_oauth1_credential = req.credential
        return req.credential

    def __call__(self, scope=None):
        def wrapper(f):
            @functools.wraps(f)
            def decorated(*args, **kwargs):
                try:
                    self.acquire_credential()
                except OAuth1Error as error:
                    body = dict(error.get_body())
                    return Response(
                        json.dumps(body),
                        status=error.status_code,
                        headers=default_json_headers,
                    )
                return f(*args, **kwargs)

            return decorated

        return wrapper


def _get_current_credential():
    return g.get("authlib_server_oauth1_credential")


current_credential = LocalProxy(_get_current_credential)



import base64

from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode


def create_basic_auth(username, password):
    text = f"{username}:{password}"
    auth = to_unicode(base64.b64encode(to_bytes(text)))
    return "Basic " + auth



import typing

import httpx
from httpx import Auth
from httpx import Request
from httpx import Response

from authlib.common.encoding import to_unicode
from authlib.oauth1 import SIGNATURE_HMAC_SHA1
from authlib.oauth1 import SIGNATURE_TYPE_HEADER
from authlib.oauth1 import ClientAuth
from authlib.oauth1.client import OAuth1Client as _OAuth1Client

from ..base_client import OAuthError
from .utils import build_request
from .utils import extract_client_kwargs


class OAuth1Auth(Auth, ClientAuth):
    """Signs the httpx request using OAuth 1 (RFC5849)."""

    requires_request_body = True

    def auth_flow(self, request: Request) -> typing.Generator[Request, Response, None]:
        url, headers, body = self.prepare(
            request.method, str(request.url), request.headers, request.content
        )
        headers["Content-Length"] = str(len(body))
        yield build_request(
            url=url, headers=headers, body=body, initial_request=request
        )


class AsyncOAuth1Client(_OAuth1Client, httpx.AsyncClient):
    auth_class = OAuth1Auth

    def __init__(
        self,
        client_id,
        client_secret=None,
        token=None,
        token_secret=None,
        redirect_uri=None,
        rsa_key=None,
        verifier=None,
        signature_method=SIGNATURE_HMAC_SHA1,
        signature_type=SIGNATURE_TYPE_HEADER,
        force_include_body=False,
        **kwargs,
    ):
        _client_kwargs = extract_client_kwargs(kwargs)
        httpx.AsyncClient.__init__(self, **_client_kwargs)

        _OAuth1Client.__init__(
            self,
            None,
            client_id=client_id,
            client_secret=client_secret,
            token=token,
            token_secret=token_secret,
            redirect_uri=redirect_uri,
            rsa_key=rsa_key,
            verifier=verifier,
            signature_method=signature_method,
            signature_type=signature_type,
            force_include_body=force_include_body,
            **kwargs,
        )

    async def fetch_access_token(self, url, verifier=None, **kwargs):
        """Method for fetching an access token from the token endpoint.

        This is the final step in the OAuth 1 workflow. An access token is
        obtained using all previously obtained credentials, including the
        verifier from the authorization step.

        :param url: Access Token endpoint.
        :param verifier: A verifier string to prove authorization was granted.
        :param kwargs: Extra parameters to include for fetching access token.
        :return: A token dict.
        """
        if verifier:
            self.auth.verifier = verifier
        if not self.auth.verifier:
            self.handle_error("missing_verifier", 'Missing "verifier" value')
        token = await self._fetch_token(url, **kwargs)
        self.auth.verifier = None
        return token

    async def _fetch_token(self, url, **kwargs):
        resp = await self.post(url, **kwargs)
        text = await resp.aread()
        token = self.parse_response_token(resp.status_code, to_unicode(text))
        self.token = token
        return token

    @staticmethod
    def handle_error(error_type, error_description):
        raise OAuthError(error_type, error_description)


class OAuth1Client(_OAuth1Client, httpx.Client):
    auth_class = OAuth1Auth

    def __init__(
        self,
        client_id,
        client_secret=None,
        token=None,
        token_secret=None,
        redirect_uri=None,
        rsa_key=None,
        verifier=None,
        signature_method=SIGNATURE_HMAC_SHA1,
        signature_type=SIGNATURE_TYPE_HEADER,
        force_include_body=False,
        **kwargs,
    ):
        _client_kwargs = extract_client_kwargs(kwargs)
        # app keyword was dropped!
        app_value = _client_kwargs.pop("app", None)
        if app_value is not None:
            _client_kwargs["transport"] = httpx.WSGITransport(app=app_value)

        httpx.Client.__init__(self, **_client_kwargs)

        _OAuth1Client.__init__(
            self,
            self,
            client_id=client_id,
            client_secret=client_secret,
            token=token,
            token_secret=token_secret,
            redirect_uri=redirect_uri,
            rsa_key=rsa_key,
            verifier=verifier,
            signature_method=signature_method,
            signature_type=signature_type,
            force_include_body=force_include_body,
            **kwargs,
        )

    @staticmethod
    def handle_error(error_type, error_description):
        raise OAuthError(error_type, error_description)



from .auth import ClientAuth
from .auth import TokenAuth
from .base import OAuth2Error
from .client import OAuth2Client
from .rfc6749 import AuthorizationServer
from .rfc6749 import ClientAuthentication
from .rfc6749 import JsonRequest
from .rfc6749 import OAuth2Request
from .rfc6749 import ResourceProtector

__all__ = [
    "OAuth2Error",
    "ClientAuth",
    "TokenAuth",
    "OAuth2Client",
    "OAuth2Request",
    "JsonRequest",
    "AuthorizationServer",
    "ClientAuthentication",
    "ResourceProtector",
]



from .authorization_server import JWTAuthenticationRequest
from .discovery import AuthorizationServerMetadata
from .registration import ClientMetadataClaims

__all__ = [
    "AuthorizationServerMetadata",
    "JWTAuthenticationRequest",
    "ClientMetadataClaims",
]



import json

import pytest

from authlib.common.urls import add_params_to_uri
from authlib.jose import JsonWebToken
from authlib.jose import jwt
from authlib.oauth2 import rfc7591
from authlib.oauth2 import rfc9101
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from tests.util import read_file_path

from .models import Client
from .models import CodeGrantMixin
from .models import save_authorization_code

authorize_url = "/oauth/authorize"


@pytest.fixture
def metadata():
    return {}


@pytest.fixture(autouse=True)
def server(server):
    class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
        TOKEN_ENDPOINT_AUTH_METHODS = [
            "client_secret_basic",
            "client_secret_post",
            "none",
        ]

        def save_authorization_code(self, code, request):
            return save_authorization_code(code, request)

    server.register_grant(AuthorizationCodeGrant)
    return server


@pytest.fixture(autouse=True)
def client_registration_endpoint(app, server, metadata, db):
    class ClientRegistrationEndpoint(rfc7591.ClientRegistrationEndpoint):
        software_statement_alg_values_supported = ["RS256"]

        def authenticate_token(self, request):
            auth_header = request.headers.get("Authorization")
            request.user_id = 1
            return auth_header

        def resolve_public_key(self, request):
            return read_file_path("rsa_public.pem")

        def save_client(self, client_info, client_metadata, request):
            client = Client(user_id=request.user_id, **client_info)
            client.set_client_metadata(client_metadata)
            db.session.add(client)
            db.session.commit()
            return client

        def get_server_metadata(self):
            return metadata

    server.register_endpoint(
        ClientRegistrationEndpoint(
            claims_classes=[
                rfc7591.ClientMetadataClaims,
                rfc9101.ClientMetadataClaims,
            ]
        )
    )

    @app.route("/create_client", methods=["POST"])
    def create_client():
        return server.create_endpoint_response(ClientRegistrationEndpoint.ENDPOINT_NAME)


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
            "jwks": read_file_path("jwks_public.json"),
            "require_signed_request_object": False,
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def register_request_object_extension(
    server,
    metadata=None,
    request_object=None,
    support_request=True,
    support_request_uri=True,
):
    class JWTAuthenticationRequest(rfc9101.JWTAuthenticationRequest):
        def resolve_client_public_key(self, client):
            return read_file_path("jwk_public.json")

        def get_request_object(self, request_uri: str):
            return request_object

        def get_server_metadata(self):
            return metadata or {}

        def get_client_require_signed_request_object(self, client):
            return client.client_metadata.get("require_signed_request_object", False)

    server.register_extension(
        JWTAuthenticationRequest(
            support_request=support_request, support_request_uri=support_request_uri
        )
    )


def test_request_parameter_get(test_client, server):
    """Pass the authentication payload in a JWT in the request query parameter."""
    register_request_object_extension(server)
    payload = {"response_type": "code", "client_id": "client-id"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    url = add_params_to_uri(
        authorize_url, {"client_id": "client-id", "request": request_obj}
    )
    rv = test_client.get(url)
    assert rv.data == b"ok"


def test_request_uri_parameter_get(test_client, server):
    """Pass the authentication payload in a JWT in the request_uri query parameter."""
    payload = {"response_type": "code", "client_id": "client-id"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    register_request_object_extension(server, request_object=request_obj)

    url = add_params_to_uri(
        authorize_url,
        {
            "client_id": "client-id",
            "request_uri": "https://client.test/request_object",
        },
    )
    rv = test_client.get(url)
    assert rv.data == b"ok"


def test_request_and_request_uri_parameters(test_client, server):
    """Passing both requests and request_uri parameters should return an error."""

    payload = {"response_type": "code", "client_id": "client-id"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    register_request_object_extension(server, request_object=request_obj)

    url = add_params_to_uri(
        authorize_url,
        {
            "client_id": "client-id",
            "request": request_obj,
            "request_uri": "https://client.test/request_object",
        },
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request"
    assert (
        params["error_description"]
        == "The 'request' and 'request_uri' parameters are mutually exclusive."
    )


def test_neither_request_nor_request_uri_parameter(test_client, server):
    """Passing parameters in the query string and not in a request object should still work."""

    register_request_object_extension(server)
    url = add_params_to_uri(
        authorize_url, {"response_type": "code", "client_id": "client-id"}
    )
    rv = test_client.get(url)
    assert rv.data == b"ok"


def test_server_require_request_object(test_client, server, metadata):
    """When server metadata 'require_signed_request_object' is true, request objects must be used."""
    metadata["require_signed_request_object"] = True
    register_request_object_extension(server, metadata=metadata)
    url = add_params_to_uri(
        authorize_url, {"response_type": "code", "client_id": "client-id"}
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request"
    assert (
        params["error_description"]
        == "Authorization requests for this server must use signed request objects."
    )


def test_server_require_request_object_alg_none(test_client, server, metadata):
    """When server metadata 'require_signed_request_object' is true, the JWT alg cannot be none."""

    metadata["require_signed_request_object"] = True
    register_request_object_extension(server, metadata=metadata)
    payload = {"response_type": "code", "client_id": "client-id"}
    jwt_none = JsonWebToken(["none"])
    request_obj = jwt_none.encode(
        {"alg": "none"}, payload, read_file_path("jwk_private.json")
    )
    url = add_params_to_uri(
        authorize_url, {"client_id": "client-id", "request": request_obj}
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request"
    assert (
        params["error_description"]
        == "Authorization requests for this server must use signed request objects."
    )


def test_client_require_signed_request_object(test_client, client, server, db):
    """When client metadata 'require_signed_request_object' is true, request objects must be used."""

    register_request_object_extension(server)
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
            "jwks": read_file_path("jwks_public.json"),
            "require_signed_request_object": True,
        }
    )
    db.session.add(client)
    db.session.commit()

    url = add_params_to_uri(
        authorize_url, {"response_type": "code", "client_id": "client-id"}
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request"
    assert (
        params["error_description"]
        == "Authorization requests for this client must use signed request objects."
    )


def test_client_require_signed_request_object_alg_none(test_client, client, server, db):
    """When client metadata 'require_signed_request_object' is true, the JWT alg cannot be none."""

    register_request_object_extension(server)
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
            "jwks": read_file_path("jwks_public.json"),
            "require_signed_request_object": True,
        }
    )
    db.session.add(client)
    db.session.commit()

    payload = {"response_type": "code", "client_id": "client-id"}
    jwt_none = JsonWebToken(["none"])
    request_obj = jwt_none.encode({"alg": "none"}, payload, "")
    url = add_params_to_uri(
        authorize_url, {"client_id": "client-id", "request": request_obj}
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request"
    assert (
        params["error_description"]
        == "Authorization requests for this client must use signed request objects."
    )


def test_unsupported_request_parameter(test_client, server):
    """Passing the request parameter when unsupported should raise a 'request_not_supported' error."""

    register_request_object_extension(server, support_request=False)
    payload = {"response_type": "code", "client_id": "client-id"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    url = add_params_to_uri(
        authorize_url, {"client_id": "client-id", "request": request_obj}
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "request_not_supported"
    assert (
        params["error_description"]
        == "The authorization server does not support the use of the request parameter."
    )


def test_unsupported_request_uri_parameter(test_client, server):
    """Passing the request parameter when unsupported should raise a 'request_uri_not_supported' error."""

    payload = {"response_type": "code", "client_id": "client-id"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    register_request_object_extension(
        server, request_object=request_obj, support_request_uri=False
    )

    url = add_params_to_uri(
        authorize_url,
        {
            "client_id": "client-id",
            "request_uri": "https://client.test/request_object",
        },
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "request_uri_not_supported"
    assert (
        params["error_description"]
        == "The authorization server does not support the use of the request_uri parameter."
    )


def test_invalid_request_uri_parameter(test_client, server):
    """Invalid request_uri (or unreachable etc.) should raise a invalid_request_uri error."""

    register_request_object_extension(server)
    url = add_params_to_uri(
        authorize_url,
        {
            "client_id": "client-id",
            "request_uri": "https://client.test/request_object",
        },
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request_uri"
    assert (
        params["error_description"]
        == "The request_uri in the authorization request returns an error or contains invalid data."
    )


def test_invalid_request_object(test_client, server):
    """Invalid request object should raise a invalid_request_object error."""

    register_request_object_extension(server)
    url = add_params_to_uri(
        authorize_url,
        {
            "client_id": "client-id",
            "request": "invalid",
        },
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request_object"
    assert (
        params["error_description"]
        == "The request parameter contains an invalid Request Object."
    )


def test_missing_client_id(test_client, server):
    """The client_id parameter is mandatory."""

    register_request_object_extension(server)
    payload = {"response_type": "code", "client_id": "client-id"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    url = add_params_to_uri(authorize_url, {"request": request_obj})

    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_client"
    assert params["error_description"] == "Missing 'client_id' parameter."


def test_invalid_client_id(test_client, server):
    """The client_id parameter is mandatory."""

    register_request_object_extension(server)
    payload = {"response_type": "code", "client_id": "invalid"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    url = add_params_to_uri(
        authorize_url, {"client_id": "invalid", "request": request_obj}
    )

    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_client"
    assert params["error_description"] == "The client does not exist on this server."


def test_different_client_id(test_client, server):
    """The client_id parameter should be the same in the request payload and the request object."""

    register_request_object_extension(server)
    payload = {"response_type": "code", "client_id": "other-code-client"}
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    url = add_params_to_uri(
        authorize_url, {"client_id": "client-id", "request": request_obj}
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request"
    assert (
        params["error_description"]
        == "The 'client_id' claim from the request parameters and the request object claims don't match."
    )


def test_request_param_in_request_object(test_client, server):
    """The request and request_uri parameters should not be present in the request object."""

    register_request_object_extension(server)
    payload = {
        "response_type": "code",
        "client_id": "client-id",
        "request_uri": "https://client.test/request_object",
    }
    request_obj = jwt.encode(
        {"alg": "RS256"}, payload, read_file_path("jwk_private.json")
    )
    url = add_params_to_uri(
        authorize_url, {"client_id": "client-id", "request": request_obj}
    )
    rv = test_client.get(url)
    params = json.loads(rv.data)
    assert params["error"] == "invalid_request"
    assert (
        params["error_description"]
        == "The 'request' and 'request_uri' parameters must not be included in the request object."
    )


def test_registration(test_client, server):
    """The 'require_signed_request_object' parameter should be available for client registration."""
    register_request_object_extension(server)
    headers = {"Authorization": "bearer abc"}

    # Default case
    body = {
        "client_name": "Authlib",
    }
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["client_name"] == "Authlib"
    assert resp["require_signed_request_object"] is False

    # Nominal case
    body = {
        "require_signed_request_object": True,
        "client_name": "Authlib",
    }
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["client_name"] == "Authlib"
    assert resp["require_signed_request_object"] is True

    # Error case
    body = {
        "require_signed_request_object": "invalid",
        "client_name": "Authlib",
    }
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client_metadata"



from authlib.common.urls import is_valid_url
from authlib.jose import BaseClaims
from authlib.jose.errors import InvalidClaimError


class ClientMetadataClaims(BaseClaims):
    REGISTERED_CLAIMS = [
        "token_endpoint_auth_signing_alg",
        "application_type",
        "sector_identifier_uri",
        "subject_type",
        "id_token_signed_response_alg",
        "id_token_encrypted_response_alg",
        "id_token_encrypted_response_enc",
        "userinfo_signed_response_alg",
        "userinfo_encrypted_response_alg",
        "userinfo_encrypted_response_enc",
        "default_max_age",
        "require_auth_time",
        "default_acr_values",
        "initiate_login_uri",
        "request_object_signing_alg",
        "request_object_encryption_alg",
        "request_object_encryption_enc",
        "request_uris",
    ]

    def validate(self):
        self._validate_essential_claims()
        self.validate_token_endpoint_auth_signing_alg()
        self.validate_application_type()
        self.validate_sector_identifier_uri()
        self.validate_subject_type()
        self.validate_id_token_signed_response_alg()
        self.validate_id_token_encrypted_response_alg()
        self.validate_id_token_encrypted_response_enc()
        self.validate_userinfo_signed_response_alg()
        self.validate_userinfo_encrypted_response_alg()
        self.validate_userinfo_encrypted_response_enc()
        self.validate_default_max_age()
        self.validate_require_auth_time()
        self.validate_default_acr_values()
        self.validate_initiate_login_uri()
        self.validate_request_object_signing_alg()
        self.validate_request_object_encryption_alg()
        self.validate_request_object_encryption_enc()
        self.validate_request_uris()

    def _validate_uri(self, key):
        uri = self.get(key)
        uris = uri if isinstance(uri, list) else [uri]
        for uri in uris:
            if uri and not is_valid_url(uri):
                raise InvalidClaimError(key)

    @classmethod
    def get_claims_options(self, metadata):
        """Generate claims options validation from Authorization Server metadata."""
        options = {}

        if acr_values_supported := metadata.get("acr_values_supported"):

            def _validate_default_acr_values(claims, value):
                return not value or set(value).issubset(set(acr_values_supported))

            options["default_acr_values"] = {"validate": _validate_default_acr_values}

        values_mapping = {
            "token_endpoint_auth_signing_alg_values_supported": "token_endpoint_auth_signing_alg",
            "subject_types_supported": "subject_type",
            "id_token_signing_alg_values_supported": "id_token_signed_response_alg",
            "id_token_encryption_alg_values_supported": "id_token_encrypted_response_alg",
            "id_token_encryption_enc_values_supported": "id_token_encrypted_response_enc",
            "userinfo_signing_alg_values_supported": "userinfo_signed_response_alg",
            "userinfo_encryption_alg_values_supported": "userinfo_encrypted_response_alg",
            "userinfo_encryption_enc_values_supported": "userinfo_encrypted_response_enc",
            "request_object_signing_alg_values_supported": "request_object_signing_alg",
            "request_object_encryption_alg_values_supported": "request_object_encryption_alg",
            "request_object_encryption_enc_values_supported": "request_object_encryption_enc",
        }

        def make_validator(metadata_claim_values):
            def _validate(claims, value):
                return not value or value in metadata_claim_values

            return _validate

        for metadata_claim_name, request_claim_name in values_mapping.items():
            if metadata_claim_values := metadata.get(metadata_claim_name):
                options[request_claim_name] = {
                    "validate": make_validator(metadata_claim_values)
                }

        return options

    def validate_token_endpoint_auth_signing_alg(self):
        """JWS [JWS] alg algorithm [JWA] that MUST be used for signing the JWT [JWT]
        used to authenticate the Client at the Token Endpoint for the private_key_jwt
        and client_secret_jwt authentication methods.

        All Token Requests using these authentication methods from this Client MUST be
        rejected, if the JWT is not signed with this algorithm. Servers SHOULD support
        RS256. The value none MUST NOT be used. The default, if omitted, is that any
        algorithm supported by the OP and the RP MAY be used.
        """
        if self.get("token_endpoint_auth_signing_alg") == "none":
            raise InvalidClaimError("token_endpoint_auth_signing_alg")

        self._validate_claim_value("token_endpoint_auth_signing_alg")

    def validate_application_type(self):
        """Kind of the application.

        The default, if omitted, is web. The defined values are native or web. Web
        Clients using the OAuth Implicit Grant Type MUST only register URLs using the
        https scheme as redirect_uris; they MUST NOT use localhost as the hostname.
        Native Clients MUST only register redirect_uris using custom URI schemes or
        loopback URLs using the http scheme; loopback URLs use localhost or the IP
        loopback literals 127.0.0.1 or [::1] as the hostname. Authorization Servers MAY
        place additional constraints on Native Clients. Authorization Servers MAY
        reject Redirection URI values using the http scheme, other than the loopback
        case for Native Clients. The Authorization Server MUST verify that all the
        registered redirect_uris conform to these constraints. This prevents sharing a
        Client ID across different types of Clients.
        """
        self.setdefault("application_type", "web")
        if self.get("application_type") not in ("web", "native"):
            raise InvalidClaimError("application_type")

        self._validate_claim_value("application_type")

    def validate_sector_identifier_uri(self):
        """URL using the https scheme to be used in calculating Pseudonymous Identifiers
        by the OP.

        The URL references a file with a single JSON array of redirect_uri values.
        Please see Section 5. Providers that use pairwise sub (subject) values SHOULD
        utilize the sector_identifier_uri value provided in the Subject Identifier
        calculation for pairwise identifiers.
        """
        self._validate_uri("sector_identifier_uri")

    def validate_subject_type(self):
        """subject_type requested for responses to this Client.

        The subject_types_supported discovery parameter contains a list of the supported
        subject_type values for the OP. Valid types include pairwise and public.
        """
        self._validate_claim_value("subject_type")

    def validate_id_token_signed_response_alg(self):
        """JWS alg algorithm [JWA] REQUIRED for signing the ID Token issued to this
        Client.

        The value none MUST NOT be used as the ID Token alg value unless the Client uses
        only Response Types that return no ID Token from the Authorization Endpoint
        (such as when only using the Authorization Code Flow). The default, if omitted,
        is RS256. The public key for validating the signature is provided by retrieving
        the JWK Set referenced by the jwks_uri element from OpenID Connect Discovery 1.0
        [OpenID.Discovery].
        """
        if self.get(
            "id_token_signed_response_alg"
        ) == "none" and "id_token" in self.get("response_type", ""):
            raise InvalidClaimError("id_token_signed_response_alg")

        self.setdefault("id_token_signed_response_alg", "RS256")
        self._validate_claim_value("id_token_signed_response_alg")

    def validate_id_token_encrypted_response_alg(self):
        """JWE alg algorithm [JWA] REQUIRED for encrypting the ID Token issued to this
        Client.

        If this is requested, the response will be signed then encrypted, with the
        result being a Nested JWT, as defined in [JWT]. The default, if omitted, is that
        no encryption is performed.
        """
        self._validate_claim_value("id_token_encrypted_response_alg")

    def validate_id_token_encrypted_response_enc(self):
        """JWE enc algorithm [JWA] REQUIRED for encrypting the ID Token issued to this
        Client.

        If id_token_encrypted_response_alg is specified, the default
        id_token_encrypted_response_enc value is A128CBC-HS256. When
        id_token_encrypted_response_enc is included, id_token_encrypted_response_alg
        MUST also be provided.
        """
        if self.get("id_token_encrypted_response_enc") and not self.get(
            "id_token_encrypted_response_alg"
        ):
            raise InvalidClaimError("id_token_encrypted_response_enc")

        if self.get("id_token_encrypted_response_alg"):
            self.setdefault("id_token_encrypted_response_enc", "A128CBC-HS256")

        self._validate_claim_value("id_token_encrypted_response_enc")

    def validate_userinfo_signed_response_alg(self):
        """JWS alg algorithm [JWA] REQUIRED for signing UserInfo Responses.

        If this is specified, the response will be JWT [JWT] serialized, and signed
        using JWS. The default, if omitted, is for the UserInfo Response to return the
        Claims as a UTF-8 [RFC3629] encoded JSON object using the application/json
        content-type.
        """
        self._validate_claim_value("userinfo_signed_response_alg")

    def validate_userinfo_encrypted_response_alg(self):
        """JWE [JWE] alg algorithm [JWA] REQUIRED for encrypting UserInfo Responses.

        If both signing and encryption are requested, the response will be signed then
        encrypted, with the result being a Nested JWT, as defined in [JWT]. The default,
        if omitted, is that no encryption is performed.
        """
        self._validate_claim_value("userinfo_encrypted_response_alg")

    def validate_userinfo_encrypted_response_enc(self):
        """JWE enc algorithm [JWA] REQUIRED for encrypting UserInfo Responses.

        If userinfo_encrypted_response_alg is specified, the default
        userinfo_encrypted_response_enc value is A128CBC-HS256. When
        userinfo_encrypted_response_enc is included, userinfo_encrypted_response_alg
        MUST also be provided.
        """
        if self.get("userinfo_encrypted_response_enc") and not self.get(
            "userinfo_encrypted_response_alg"
        ):
            raise InvalidClaimError("userinfo_encrypted_response_enc")

        if self.get("userinfo_encrypted_response_alg"):
            self.setdefault("userinfo_encrypted_response_enc", "A128CBC-HS256")

        self._validate_claim_value("userinfo_encrypted_response_enc")

    def validate_default_max_age(self):
        """Default Maximum Authentication Age.

        Specifies that the End-User MUST be actively authenticated if the End-User was
        authenticated longer ago than the specified number of seconds. The max_age
        request parameter overrides this default value. If omitted, no default Maximum
        Authentication Age is specified.
        """
        if self.get("default_max_age") is not None and not isinstance(
            self["default_max_age"], (int, float)
        ):
            raise InvalidClaimError("default_max_age")

        self._validate_claim_value("default_max_age")

    def validate_require_auth_time(self):
        """Boolean value specifying whether the auth_time Claim in the ID Token is
        REQUIRED.

        It is REQUIRED when the value is true. (If this is false, the auth_time Claim
        can still be dynamically requested as an individual Claim for the ID Token using
        the claims request parameter described in Section 5.5.1 of OpenID Connect Core
        1.0 [OpenID.Core].) If omitted, the default value is false.
        """
        self.setdefault("require_auth_time", False)
        if self.get("require_auth_time") is not None and not isinstance(
            self["require_auth_time"], bool
        ):
            raise InvalidClaimError("require_auth_time")

        self._validate_claim_value("require_auth_time")

    def validate_default_acr_values(self):
        """Default requested Authentication Context Class Reference values.

        Array of strings that specifies the default acr values that the OP is being
        requested to use for processing requests from this Client, with the values
        appearing in order of preference. The Authentication Context Class satisfied by
        the authentication performed is returned as the acr Claim Value in the issued ID
        Token. The acr Claim is requested as a Voluntary Claim by this parameter. The
        acr_values_supported discovery element contains a list of the supported acr
        values supported by the OP. Values specified in the acr_values request parameter
        or an individual acr Claim request override these default values.
        """
        self._validate_claim_value("default_acr_values")

    def validate_initiate_login_uri(self):
        """RI using the https scheme that a third party can use to initiate a login by
        the RP, as specified in Section 4 of OpenID Connect Core 1.0 [OpenID.Core].

        The URI MUST accept requests via both GET and POST. The Client MUST understand
        the login_hint and iss parameters and SHOULD support the target_link_uri
        parameter.
        """
        self._validate_uri("initiate_login_uri")

    def validate_request_object_signing_alg(self):
        """JWS [JWS] alg algorithm [JWA] that MUST be used for signing Request Objects
        sent to the OP.

        All Request Objects from this Client MUST be rejected, if not signed with this
        algorithm. Request Objects are described in Section 6.1 of OpenID Connect Core
        1.0 [OpenID.Core]. This algorithm MUST be used both when the Request Object is
        passed by value (using the request parameter) and when it is passed by reference
        (using the request_uri parameter). Servers SHOULD support RS256. The value none
        MAY be used. The default, if omitted, is that any algorithm supported by the OP
        and the RP MAY be used.
        """
        self._validate_claim_value("request_object_signing_alg")

    def validate_request_object_encryption_alg(self):
        """JWE [JWE] alg algorithm [JWA] the RP is declaring that it may use for
        encrypting Request Objects sent to the OP.

        This parameter SHOULD be included when symmetric encryption will be used, since
        this signals to the OP that a client_secret value needs to be returned from
        which the symmetric key will be derived, that might not otherwise be returned.
        The RP MAY still use other supported encryption algorithms or send unencrypted
        Request Objects, even when this parameter is present. If both signing and
        encryption are requested, the Request Object will be signed then encrypted, with
        the result being a Nested JWT, as defined in [JWT]. The default, if omitted, is
        that the RP is not declaring whether it might encrypt any Request Objects.
        """
        self._validate_claim_value("request_object_encryption_alg")

    def validate_request_object_encryption_enc(self):
        """JWE enc algorithm [JWA] the RP is declaring that it may use for encrypting
        Request Objects sent to the OP.

        If request_object_encryption_alg is specified, the default
        request_object_encryption_enc value is A128CBC-HS256. When
        request_object_encryption_enc is included, request_object_encryption_alg MUST
        also be provided.
        """
        if self.get("request_object_encryption_enc") and not self.get(
            "request_object_encryption_alg"
        ):
            raise InvalidClaimError("request_object_encryption_enc")

        if self.get("request_object_encryption_alg"):
            self.setdefault("request_object_encryption_enc", "A128CBC-HS256")

        self._validate_claim_value("request_object_encryption_enc")

    def validate_request_uris(self):
        """Array of request_uri values that are pre-registered by the RP for use at the
        OP.

        These URLs MUST use the https scheme unless the target Request Object is signed
        in a way that is verifiable by the OP. Servers MAY cache the contents of the
        files referenced by these URIs and not retrieve them at the time they are used
        in a request. OPs can require that request_uri values used be pre-registered
        with the require_request_uri_registration discovery parameter. If the contents
        of the request file could ever change, these URI values SHOULD include the
        base64url-encoded SHA-256 hash value of the file contents referenced by the URI
        as the value of the URI fragment. If the fragment value used for a URI changes,
        that signals the server that its cached value for that URI with the old fragment
        value is no longer valid.
        """
        self._validate_uri("request_uris")



import zlib

from ..rfc7516 import JsonWebEncryption
from ..rfc7516 import JWEZipAlgorithm

GZIP_HEAD = bytes([120, 156])
MAX_SIZE = 250 * 1024


class DeflateZipAlgorithm(JWEZipAlgorithm):
    name = "DEF"
    description = "DEFLATE"

    def compress(self, s: bytes) -> bytes:
        """Compress bytes data with DEFLATE algorithm."""
        data = zlib.compress(s)
        # https://datatracker.ietf.org/doc/html/rfc1951
        # since DEF is always gzip, we can drop gzip headers and tail
        return data[2:-4]

    def decompress(self, s: bytes) -> bytes:
        """Decompress DEFLATE bytes data."""
        if s.startswith(GZIP_HEAD):
            decompressor = zlib.decompressobj()
        else:
            decompressor = zlib.decompressobj(-zlib.MAX_WBITS)
        value = decompressor.decompress(s, MAX_SIZE)
        if decompressor.unconsumed_tail:
            raise ValueError(f"Decompressed string exceeds {MAX_SIZE} bytes")
        return value


def register_jwe_rfc7518():
    JsonWebEncryption.register_algorithm(DeflateZipAlgorithm())



# flake8: noqa

from .authorization_server import AuthorizationServer
from .endpoints import RevocationEndpoint
from .resource_protector import BearerTokenValidator
from .resource_protector import ResourceProtector
from .signals import client_authenticated
from .signals import token_authenticated
from .signals import token_revoked



from requests import Session
from requests.auth import AuthBase

from authlib.common.encoding import to_native
from authlib.oauth1 import SIGNATURE_HMAC_SHA1
from authlib.oauth1 import SIGNATURE_TYPE_HEADER
from authlib.oauth1 import ClientAuth
from authlib.oauth1.client import OAuth1Client

from ..base_client import OAuthError
from .utils import update_session_configure


class OAuth1Auth(AuthBase, ClientAuth):
    """Signs the request using OAuth 1 (RFC5849)."""

    def __call__(self, req):
        url, headers, body = self.prepare(req.method, req.url, req.headers, req.body)

        req.url = to_native(url)
        req.prepare_headers(headers)
        if body:
            req.body = body
        return req


class OAuth1Session(OAuth1Client, Session):
    auth_class = OAuth1Auth

    def __init__(
        self,
        client_id,
        client_secret=None,
        token=None,
        token_secret=None,
        redirect_uri=None,
        rsa_key=None,
        verifier=None,
        signature_method=SIGNATURE_HMAC_SHA1,
        signature_type=SIGNATURE_TYPE_HEADER,
        force_include_body=False,
        **kwargs,
    ):
        Session.__init__(self)
        update_session_configure(self, kwargs)
        OAuth1Client.__init__(
            self,
            session=self,
            client_id=client_id,
            client_secret=client_secret,
            token=token,
            token_secret=token_secret,
            redirect_uri=redirect_uri,
            rsa_key=rsa_key,
            verifier=verifier,
            signature_method=signature_method,
            signature_type=signature_type,
            force_include_body=force_include_body,
            **kwargs,
        )

    def rebuild_auth(self, prepared_request, response):
        """When being redirected we should always strip Authorization
        header, since nonce may not be reused as per OAuth spec.
        """
        if "Authorization" in prepared_request.headers:
            # If we get redirected to a new host, we should strip out
            # any authentication headers.
            prepared_request.headers.pop("Authorization", True)
            prepared_request.prepare_auth(self.auth)

    @staticmethod
    def handle_error(error_type, error_description):
        raise OAuthError(error_type, error_description)



import json

import pytest

from authlib.jose import JsonWebSignature
from authlib.jose import errors
from tests.util import read_file_path


def test_invalid_input():
    jws = JsonWebSignature()
    with pytest.raises(errors.DecodeError):
        jws.deserialize("a", "k")
    with pytest.raises(errors.DecodeError):
        jws.deserialize("a.b.c", "k")
    with pytest.raises(errors.DecodeError):
        jws.deserialize("YQ.YQ.YQ", "k")  # a
    with pytest.raises(errors.DecodeError):
        jws.deserialize("W10.a.YQ", "k")  # []
    with pytest.raises(errors.DecodeError):
        jws.deserialize("e30.a.YQ", "k")  # {}
    with pytest.raises(errors.DecodeError):
        jws.deserialize("eyJhbGciOiJzIn0.a.YQ", "k")
    with pytest.raises(errors.DecodeError):
        jws.deserialize("eyJhbGciOiJzIn0.YQ.a", "k")


def test_invalid_alg():
    jws = JsonWebSignature()
    with pytest.raises(errors.UnsupportedAlgorithmError):
        jws.deserialize(
            "eyJhbGciOiJzIn0.YQ.YQ",
            "k",
        )
    with pytest.raises(errors.MissingAlgorithmError):
        jws.serialize({}, "", "k")
    with pytest.raises(errors.UnsupportedAlgorithmError):
        jws.serialize({"alg": "s"}, "", "k")


def test_bad_signature():
    jws = JsonWebSignature()
    s = "eyJhbGciOiJIUzI1NiJ9.YQ.YQ"
    with pytest.raises(errors.BadSignatureError):
        jws.deserialize(s, "k")


def test_not_supported_alg():
    jws = JsonWebSignature(algorithms=["HS256"])
    s = jws.serialize({"alg": "HS256"}, "hello", "secret")

    jws = JsonWebSignature(algorithms=["RS256"])
    with pytest.raises(errors.UnsupportedAlgorithmError):
        jws.serialize({"alg": "HS256"}, "hello", "secret")

    with pytest.raises(errors.UnsupportedAlgorithmError):
        jws.deserialize(s, "secret")


def test_compact_jws():
    jws = JsonWebSignature(algorithms=["HS256"])
    s = jws.serialize({"alg": "HS256"}, "hello", "secret")
    data = jws.deserialize(s, "secret")
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "HS256"
    assert "signature" not in data


def test_compact_rsa():
    jws = JsonWebSignature()
    private_key = read_file_path("rsa_private.pem")
    public_key = read_file_path("rsa_public.pem")
    s = jws.serialize({"alg": "RS256"}, "hello", private_key)
    data = jws.deserialize(s, public_key)
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "RS256"

    # can deserialize with private key
    data2 = jws.deserialize(s, private_key)
    assert data == data2

    ssh_pub_key = read_file_path("ssh_public.pem")
    with pytest.raises(errors.BadSignatureError):
        jws.deserialize(s, ssh_pub_key)


def test_compact_rsa_pss():
    jws = JsonWebSignature()
    private_key = read_file_path("rsa_private.pem")
    public_key = read_file_path("rsa_public.pem")
    s = jws.serialize({"alg": "PS256"}, "hello", private_key)
    data = jws.deserialize(s, public_key)
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "PS256"
    ssh_pub_key = read_file_path("ssh_public.pem")
    with pytest.raises(errors.BadSignatureError):
        jws.deserialize(s, ssh_pub_key)


def test_compact_none():
    jws = JsonWebSignature(algorithms=["none"])
    s = jws.serialize({"alg": "none"}, "hello", None)
    data = jws.deserialize(s, None)
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "none"


def test_flattened_json_jws():
    jws = JsonWebSignature()
    protected = {"alg": "HS256"}
    header = {"protected": protected, "header": {"kid": "a"}}
    s = jws.serialize(header, "hello", "secret")
    assert isinstance(s, dict)

    data = jws.deserialize(s, "secret")
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "HS256"
    assert "protected" not in data


def test_nested_json_jws():
    jws = JsonWebSignature()
    protected = {"alg": "HS256"}
    header = {"protected": protected, "header": {"kid": "a"}}
    s = jws.serialize([header], "hello", "secret")
    assert isinstance(s, dict)
    assert "signatures" in s

    data = jws.deserialize(s, "secret")
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header[0]["alg"] == "HS256"
    assert "signatures" not in data

    # test bad signature
    with pytest.raises(errors.BadSignatureError):
        jws.deserialize(s, "f")


def test_function_key():
    protected = {"alg": "HS256"}
    header = [
        {"protected": protected, "header": {"kid": "a"}},
        {"protected": protected, "header": {"kid": "b"}},
    ]

    def load_key(header, payload):
        assert payload == b"hello"
        kid = header.get("kid")
        if kid == "a":
            return "secret-a"
        return "secret-b"

    jws = JsonWebSignature()
    s = jws.serialize(header, b"hello", load_key)
    assert isinstance(s, dict)
    assert "signatures" in s

    data = jws.deserialize(json.dumps(s), load_key)
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header[0]["alg"] == "HS256"
    assert "signature" not in data


def test_serialize_json_empty_payload():
    jws = JsonWebSignature()
    protected = {"alg": "HS256"}
    header = {"protected": protected, "header": {"kid": "a"}}
    s = jws.serialize_json(header, b"", "secret")
    data = jws.deserialize_json(s, "secret")
    assert data["payload"] == b""


def test_fail_deserialize_json():
    jws = JsonWebSignature()
    with pytest.raises(errors.DecodeError):
        jws.deserialize_json(None, "")
    with pytest.raises(errors.DecodeError):
        jws.deserialize_json("[]", "")
    with pytest.raises(errors.DecodeError):
        jws.deserialize_json("{}", "")

    # missing protected
    s = json.dumps({"payload": "YQ"})
    with pytest.raises(errors.DecodeError):
        jws.deserialize_json(s, "")

    # missing signature
    s = json.dumps({"payload": "YQ", "protected": "YQ"})
    with pytest.raises(errors.DecodeError):
        jws.deserialize_json(s, "")


def test_serialize_json_overwrite_header():
    jws = JsonWebSignature()
    protected = {"alg": "HS256", "kid": "a"}
    header = {"protected": protected}
    result = jws.serialize_json(header, b"", "secret")
    result["header"] = {"kid": "b"}
    decoded = jws.deserialize_json(result, "secret")
    assert decoded["header"]["kid"] == "a"


def test_validate_header():
    jws = JsonWebSignature(private_headers=[])
    protected = {"alg": "HS256", "invalid": "k"}
    header = {"protected": protected, "header": {"kid": "a"}}
    with pytest.raises(errors.InvalidHeaderParameterNameError):
        jws.serialize(
            header,
            b"hello",
            "secret",
        )
    jws = JsonWebSignature(private_headers=["invalid"])
    s = jws.serialize(header, b"hello", "secret")
    assert isinstance(s, dict)

    jws = JsonWebSignature()
    s = jws.serialize(header, b"hello", "secret")
    assert isinstance(s, dict)


def test_validate_crit_header_with_serialize():
    jws = JsonWebSignature()
    protected = {"alg": "HS256", "kid": "1", "crit": ["kid"]}
    jws.serialize(protected, b"hello", "secret")

    protected = {"alg": "HS256", "crit": ["kid"]}
    with pytest.raises(errors.InvalidCritHeaderParameterNameError):
        jws.serialize(protected, b"hello", "secret")

    protected = {"alg": "HS256", "invalid": "1", "crit": ["invalid"]}
    with pytest.raises(errors.InvalidCritHeaderParameterNameError):
        jws.serialize(protected, b"hello", "secret")


def test_validate_crit_header_with_deserialize():
    jws = JsonWebSignature()
    case1 = "eyJhbGciOiJIUzI1NiIsImNyaXQiOlsia2lkIl19.aGVsbG8.RVimhJH2LRGAeHy0ZcbR9xsgKhzhxIBkHs7S_TDgWvc"
    with pytest.raises(errors.InvalidCritHeaderParameterNameError):
        jws.deserialize(case1, "secret")

    case2 = (
        "eyJhbGciOiJIUzI1NiIsImludmFsaWQiOiIxIiwiY3JpdCI6WyJpbnZhbGlkIl19."
        "aGVsbG8.ifW_D1AQWzggrpd8npcnmpiwMD9dp5FTX66lCkYFENM"
    )
    with pytest.raises(errors.InvalidCritHeaderParameterNameError):
        jws.deserialize(case2, "secret")


def test_unprotected_crit_rejected_in_json_serialize():
    jws = JsonWebSignature()
    protected = {"alg": "HS256", "kid": "a"}
    # Place 'crit' in unprotected header; must be rejected
    header = {"protected": protected, "header": {"kid": "a", "crit": ["kid"]}}
    with pytest.raises(errors.InvalidHeaderParameterNameError):
        jws.serialize_json(header, b"hello", "secret")


def test_unprotected_crit_rejected_in_json_deserialize():
    jws = JsonWebSignature()
    protected = {"alg": "HS256", "kid": "a"}
    header = {"protected": protected, "header": {"kid": "a"}}
    data = jws.serialize_json(header, b"hello", "secret")
    # Tamper by adding 'crit' into the unprotected header; must be rejected
    data_tampered = dict(data)
    data_tampered["header"] = {"kid": "a", "crit": ["kid"]}
    with pytest.raises(errors.InvalidHeaderParameterNameError):
        jws.deserialize_json(data_tampered, "secret")


def test_ES512_alg():
    jws = JsonWebSignature()
    private_key = read_file_path("secp521r1-private.json")
    public_key = read_file_path("secp521r1-public.json")
    with pytest.raises(ValueError):
        jws.serialize({"alg": "ES256"}, "hello", private_key)
    s = jws.serialize({"alg": "ES512"}, "hello", private_key)
    data = jws.deserialize(s, public_key)
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "ES512"


def test_ES256K_alg():
    jws = JsonWebSignature(algorithms=["ES256K"])
    private_key = read_file_path("secp256k1-private.pem")
    public_key = read_file_path("secp256k1-pub.pem")
    s = jws.serialize({"alg": "ES256K"}, "hello", private_key)
    data = jws.deserialize(s, public_key)
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "ES256K"


def test_deserialize_exceeds_length():
    jws = JsonWebSignature()
    value = "aa" * 256000

    # header exceeds length
    with pytest.raises(ValueError):
        jws.deserialize(value + "." + value + "." + value, "")

    # payload exceeds length
    with pytest.raises(ValueError):
        jws.deserialize("eyJhbGciOiJIUzI1NiJ9." + value + "." + value, "")

    # signature exceeds length
    with pytest.raises(ValueError):
        jws.deserialize("eyJhbGciOiJIUzI1NiJ9.YQ." + value, "")



from authlib.jose.errors import InvalidClaimError
from authlib.jose.rfc7519 import JWTClaims


class JWTAccessTokenClaims(JWTClaims):
    REGISTERED_CLAIMS = JWTClaims.REGISTERED_CLAIMS + [
        "client_id",
        "auth_time",
        "acr",
        "amr",
        "scope",
        "groups",
        "roles",
        "entitlements",
    ]

    def validate(self, **kwargs):
        self.validate_typ()

        super().validate(**kwargs)
        self.validate_client_id()
        self.validate_auth_time()
        self.validate_acr()
        self.validate_amr()
        self.validate_scope()
        self.validate_groups()
        self.validate_roles()
        self.validate_entitlements()

    def validate_typ(self):
        # The resource server MUST verify that the 'typ' header value is 'at+jwt'
        # or 'application/at+jwt' and reject tokens carrying any other value.
        # 'typ' is not a required claim, so we don't raise an error if it's missing.
        typ = self.header.get("typ")
        if typ and typ.lower() not in ("at+jwt", "application/at+jwt"):
            raise InvalidClaimError("typ")

    def validate_client_id(self):
        return self._validate_claim_value("client_id")

    def validate_auth_time(self):
        auth_time = self.get("auth_time")
        if auth_time and not isinstance(auth_time, (int, float)):
            raise InvalidClaimError("auth_time")

    def validate_acr(self):
        return self._validate_claim_value("acr")

    def validate_amr(self):
        amr = self.get("amr")
        if amr and not isinstance(self["amr"], list):
            raise InvalidClaimError("amr")

    def validate_scope(self):
        return self._validate_claim_value("scope")

    def validate_groups(self):
        return self._validate_claim_value("groups")

    def validate_roles(self):
        return self._validate_claim_value("roles")

    def validate_entitlements(self):
        return self._validate_claim_value("entitlements")



import datetime

import pytest

from authlib.jose import JsonWebKey
from authlib.jose import JsonWebToken
from authlib.jose import JWTClaims
from authlib.jose import errors
from authlib.jose import jwt
from authlib.jose.errors import UnsupportedAlgorithmError
from tests.util import read_file_path


def test_init_algorithms():
    _jwt = JsonWebToken(["RS256"])
    with pytest.raises(UnsupportedAlgorithmError):
        _jwt.encode({"alg": "HS256"}, {}, "k")

    _jwt = JsonWebToken("RS256")
    with pytest.raises(UnsupportedAlgorithmError):
        _jwt.encode({"alg": "HS256"}, {}, "k")


def test_encode_sensitive_data():
    # check=False won't raise error
    jwt.encode({"alg": "HS256"}, {"password": ""}, "k", check=False)
    with pytest.raises(errors.InsecureClaimError):
        jwt.encode(
            {"alg": "HS256"},
            {"password": ""},
            "k",
        )
    with pytest.raises(errors.InsecureClaimError):
        jwt.encode(
            {"alg": "HS256"},
            {"text": "4242424242424242"},
            "k",
        )


def test_encode_datetime():
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    id_token = jwt.encode({"alg": "HS256"}, {"exp": now}, "k")
    claims = jwt.decode(id_token, "k")
    assert isinstance(claims.exp, int)


def test_validate_essential_claims():
    id_token = jwt.encode({"alg": "HS256"}, {"iss": "foo"}, "k")
    claims_options = {"iss": {"essential": True, "values": ["foo"]}}
    claims = jwt.decode(id_token, "k", claims_options=claims_options)
    claims.validate()

    claims.options = {"sub": {"essential": True}}
    with pytest.raises(errors.MissingClaimError):
        claims.validate()


def test_attribute_error():
    claims = JWTClaims({"iss": "foo"}, {"alg": "HS256"})
    with pytest.raises(AttributeError):
        claims.invalid  # noqa: B018


def test_invalid_values():
    id_token = jwt.encode({"alg": "HS256"}, {"iss": "foo"}, "k")
    claims_options = {"iss": {"values": ["bar"]}}
    claims = jwt.decode(id_token, "k", claims_options=claims_options)
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()
    claims.options = {"iss": {"value": "bar"}}
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()


def test_validate_expected_issuer_received_None():
    id_token = jwt.encode({"alg": "HS256"}, {"iss": None, "sub": None}, "k")
    claims_options = {"iss": {"essential": True, "values": ["foo"]}}
    claims = jwt.decode(id_token, "k", claims_options=claims_options)
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()


def test_validate_aud():
    id_token = jwt.encode({"alg": "HS256"}, {"aud": "foo"}, "k")
    claims_options = {"aud": {"essential": True, "value": "foo"}}
    claims = jwt.decode(id_token, "k", claims_options=claims_options)
    claims.validate()

    claims.options = {"aud": {"values": ["bar"]}}
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()

    id_token = jwt.encode({"alg": "HS256"}, {"aud": ["foo", "bar"]}, "k")
    claims = jwt.decode(id_token, "k", claims_options=claims_options)
    claims.validate()
    # no validate
    claims.options = {"aud": {"values": []}}
    claims.validate()


def test_validate_exp():
    id_token = jwt.encode({"alg": "HS256"}, {"exp": "invalid"}, "k")
    claims = jwt.decode(id_token, "k")
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()

    id_token = jwt.encode({"alg": "HS256"}, {"exp": 1234}, "k")
    claims = jwt.decode(id_token, "k")
    with pytest.raises(errors.ExpiredTokenError):
        claims.validate()


def test_validate_nbf():
    id_token = jwt.encode({"alg": "HS256"}, {"nbf": "invalid"}, "k")
    claims = jwt.decode(id_token, "k")
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()

    id_token = jwt.encode({"alg": "HS256"}, {"nbf": 1234}, "k")
    claims = jwt.decode(id_token, "k")
    claims.validate()

    id_token = jwt.encode({"alg": "HS256"}, {"nbf": 1234}, "k")
    claims = jwt.decode(id_token, "k")
    with pytest.raises(errors.InvalidTokenError):
        claims.validate(123)


def test_validate_iat_issued_in_future():
    in_future = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(
        seconds=10
    )
    id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
    claims = jwt.decode(id_token, "k")
    with pytest.raises(
        errors.InvalidTokenError,
        match="The token is not valid as it was issued in the future",
    ):
        claims.validate()


def test_validate_iat_issued_in_future_with_insufficient_leeway():
    in_future = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(
        seconds=10
    )
    id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
    claims = jwt.decode(id_token, "k")
    with pytest.raises(
        errors.InvalidTokenError,
        match="The token is not valid as it was issued in the future",
    ):
        claims.validate(leeway=5)


def test_validate_iat_issued_in_future_with_sufficient_leeway():
    in_future = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(
        seconds=10
    )
    id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
    claims = jwt.decode(id_token, "k")
    claims.validate(leeway=20)


def test_validate_iat_issued_in_past():
    in_future = datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(
        seconds=10
    )
    id_token = jwt.encode({"alg": "HS256"}, {"iat": in_future}, "k")
    claims = jwt.decode(id_token, "k")
    claims.validate()


def test_validate_iat():
    id_token = jwt.encode({"alg": "HS256"}, {"iat": "invalid"}, "k")
    claims = jwt.decode(id_token, "k")
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()


def test_validate_jti():
    id_token = jwt.encode({"alg": "HS256"}, {"jti": "bar"}, "k")
    claims_options = {"jti": {"validate": lambda c, o: o == "foo"}}
    claims = jwt.decode(id_token, "k", claims_options=claims_options)
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()


def test_validate_custom():
    id_token = jwt.encode({"alg": "HS256"}, {"custom": "foo"}, "k")
    claims_options = {"custom": {"validate": lambda c, o: o == "bar"}}
    claims = jwt.decode(id_token, "k", claims_options=claims_options)
    with pytest.raises(errors.InvalidClaimError):
        claims.validate()


def test_use_jws():
    payload = {"name": "hi"}
    private_key = read_file_path("rsa_private.pem")
    pub_key = read_file_path("rsa_public.pem")
    data = jwt.encode({"alg": "RS256"}, payload, private_key)
    assert data.count(b".") == 2

    claims = jwt.decode(data, pub_key)
    assert claims["name"] == "hi"


def test_use_jwe():
    payload = {"name": "hi"}
    private_key = read_file_path("rsa_private.pem")
    pub_key = read_file_path("rsa_public.pem")
    _jwt = JsonWebToken(["RSA-OAEP", "A256GCM"])
    data = _jwt.encode({"alg": "RSA-OAEP", "enc": "A256GCM"}, payload, pub_key)
    assert data.count(b".") == 4

    claims = _jwt.decode(data, private_key)
    assert claims["name"] == "hi"


def test_use_jwks():
    header = {"alg": "RS256", "kid": "abc"}
    payload = {"name": "hi"}
    private_key = read_file_path("jwks_private.json")
    pub_key = read_file_path("jwks_public.json")
    data = jwt.encode(header, payload, private_key)
    assert data.count(b".") == 2
    claims = jwt.decode(data, pub_key)
    assert claims["name"] == "hi"


def test_use_jwks_single_kid():
    """Test that jwks can be decoded if a kid for decoding is given and encoded data has no kid and only one key is set."""
    header = {"alg": "RS256"}
    payload = {"name": "hi"}
    private_key = read_file_path("jwks_single_private.json")
    pub_key = read_file_path("jwks_single_public.json")
    data = jwt.encode(header, payload, private_key)
    assert data.count(b".") == 2
    claims = jwt.decode(data, pub_key)
    assert claims["name"] == "hi"


# Added a unit test to showcase my problem.
# This calls jwt.decode similarly as is done in parse_id_token method of the AsyncOpenIDMixin class when the id token does not contain a kid in the alg header.
def test_use_jwks_single_kid_keyset():
    """Test that jwks can be decoded if a kid for decoding is given and encoded data has no kid and a keyset with one key."""
    header = {"alg": "RS256"}
    payload = {"name": "hi"}
    private_key = read_file_path("jwks_single_private.json")
    pub_key = read_file_path("jwks_single_public.json")
    data = jwt.encode(header, payload, private_key)
    assert data.count(b".") == 2
    claims = jwt.decode(data, JsonWebKey.import_key_set(pub_key))
    assert claims["name"] == "hi"


def test_with_ec():
    payload = {"name": "hi"}
    private_key = read_file_path("secp521r1-private.json")
    pub_key = read_file_path("secp521r1-public.json")
    data = jwt.encode({"alg": "ES512"}, payload, private_key)
    assert data.count(b".") == 2

    claims = jwt.decode(data, pub_key)
    assert claims["name"] == "hi"



from requests import Session

from authlib.oauth2.rfc7521 import AssertionClient
from authlib.oauth2.rfc7523 import JWTBearerGrant

from .oauth2_session import OAuth2Auth
from .utils import update_session_configure


class AssertionAuth(OAuth2Auth):
    def ensure_active_token(self):
        if self.client and (
            not self.token or self.token.is_expired(self.client.leeway)
        ):
            return self.client.refresh_token()


class AssertionSession(AssertionClient, Session):
    """Constructs a new Assertion Framework for OAuth 2.0 Authorization Grants
    per RFC7521_.

    .. _RFC7521: https://tools.ietf.org/html/rfc7521
    """

    token_auth_class = AssertionAuth
    JWT_BEARER_GRANT_TYPE = JWTBearerGrant.GRANT_TYPE
    ASSERTION_METHODS = {
        JWT_BEARER_GRANT_TYPE: JWTBearerGrant.sign,
    }
    DEFAULT_GRANT_TYPE = JWT_BEARER_GRANT_TYPE

    def __init__(
        self,
        token_endpoint,
        issuer,
        subject,
        audience=None,
        grant_type=None,
        claims=None,
        token_placement="header",
        scope=None,
        default_timeout=None,
        leeway=60,
        **kwargs,
    ):
        Session.__init__(self)
        self.default_timeout = default_timeout
        update_session_configure(self, kwargs)
        AssertionClient.__init__(
            self,
            session=self,
            token_endpoint=token_endpoint,
            issuer=issuer,
            subject=subject,
            audience=audience,
            grant_type=grant_type,
            claims=claims,
            token_placement=token_placement,
            scope=scope,
            leeway=leeway,
            **kwargs,
        )

    def request(self, method, url, withhold_token=False, auth=None, **kwargs):
        """Send request with auto refresh token feature."""
        if self.default_timeout:
            kwargs.setdefault("timeout", self.default_timeout)
        if not withhold_token and auth is None:
            auth = self.token_auth
        return super().request(method, url, auth=auth, **kwargs)



import time

import pytest
from flask import json

from authlib.oauth2.rfc6749.grants import RefreshTokenGrant as _RefreshTokenGrant

from .models import Token
from .models import User
from .models import db
from .oauth2_server import create_basic_header


@pytest.fixture(autouse=True)
def server(server):
    class RefreshTokenGrant(_RefreshTokenGrant):
        def authenticate_refresh_token(self, refresh_token):
            item = Token.query.filter_by(refresh_token=refresh_token).first()
            if item and item.is_refresh_token_active():
                return item

        def authenticate_user(self, credential):
            return db.session.get(User, credential.user_id)

        def revoke_old_credential(self, credential):
            now = int(time.time())
            credential.access_token_revoked_at = now
            credential.refresh_token_revoked_at = now
            db.session.add(credential)
            db.session.commit()

    server.register_grant(RefreshTokenGrant)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "grant_types": ["refresh_token"],
            "redirect_uris": ["https://client.test/authorized"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_invalid_client(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "foo",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("invalid-client", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "foo",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("client-id", "invalid-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "foo",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_invalid_refresh_token(test_client):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"
    assert "Missing" in resp["error_description"]

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "foo",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_grant"


def test_invalid_scope(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "invalid",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_scope"


def test_invalid_scope_none(test_client, token):
    token.scope = None
    db.session.add(token)
    db.session.commit()

    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "invalid",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_scope"


def test_invalid_user(test_client, token):
    token.user_id = 5
    db.session.add(token)
    db.session.commit()

    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "profile",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"


def test_invalid_grant_type(test_client, client, db, token):
    client.set_client_metadata(
        {
            "scope": "profile",
            "grant_types": ["invalid"],
            "redirect_uris": ["https://client.test/authorized"],
        }
    )
    db.session.add(client)
    db.session.commit()

    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "profile",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unauthorized_client"


def test_authorize_token_no_scope(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_authorize_token_scope(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "profile",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_revoke_old_credential(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "profile",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "profile",
        },
        headers=headers,
    )
    assert rv.status_code == 400
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_grant"


def test_token_generator(test_client, token, app, server):
    m = "tests.flask.test_oauth2.oauth2_server:token_generator"
    app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
    server.load_config(app.config)

    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "c-refresh_token.1." in resp["access_token"]
